/***************************************************************************
   log_xml.c            Sonar standard XML file output
                             -------------------
    begin                : Thurs Sept 05 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: log_xml.c,v 1.13 2004/04/24 06:22:47 cappaberra Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file log_xml.c
  * Contains the XML log file output plugin
  */

#include <assert.h>
#include "log_xml.h"
#include "mem.h"

static const char rcsid[] =
    "\100$ autosec/sonar: $Id: log_xml.c,v 1.13 2004/04/24 06:22:47 cappaberra Exp $";

/// Plugin's plugin_t structure, built by plugin_init
static plugin_t pthis;
/// Sonar's sonar_t structure, built by sonar, passed to this plugin
static sonar_t *mthis;
/// filename
static char *my_file;
/// file pointer
static FILE *my_fp;

/// my last error
static uint16_t my_errno;

static int plugin_destroy(void);
static int plugin_open(void *c_data, int i_data);
static int plugin_close(void);
static int plugin_output(char *msg, va_list * ap);
static int plugin_results(target_t * result, unsigned int num);
static void plugin_usage(void);
static void plugin_status(void);
static int plugin_last_error(void);

static int header = 0;
/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */

void *
plugin_init(void *in_data)
{
    assert(in_data != NULL);

    // useless reference, to avoid a warning
    if(rcsid) ;

    // this is assumed
    mthis = (sonar_t *) in_data;
    if(mthis->api < 6)
        mthis->err_ptrs->err_quit("Plugin API must be 6 or higher, update sonar");

    pthis.type = OUTPUT_PLUGIN;
    pthis.err = &my_errno;
    pthis.ready = false;
    pthis.version = 6;          /* set API version 6 */

    pthis.plugin_char = 'X';
    pthis.shortname = "xml";
    pthis.longname = "XML Log File output plugin";

    pthis.plugin_init = &plugin_init;
    pthis.plugin_destroy = &plugin_destroy;
    pthis.plugin_open = &plugin_open;
    pthis.plugin_close = &plugin_close;
    pthis.plugin_next = NULL;
    // this is not needed for I/O plugins
    pthis.plugin_sweep = NULL;
    pthis.plugin_output = &plugin_output;
    pthis.plugin_results = &plugin_results;
    pthis.plugin_usage = &plugin_usage;
    pthis.plugin_status = &plugin_status;
    pthis.plugin_last_error = &plugin_last_error;
    my_errno = PLUGIN_LOADED;
    my_file = NULL;
    my_fp = NULL;

    return &pthis;
}


static int
plugin_destroy(void)
{
    mthis = NULL;

    pthis.type = 0;
    pthis.err = NULL;
    pthis.ready = false;
    pthis.plugin_init = NULL;
    pthis.plugin_destroy = NULL;
    pthis.plugin_open = NULL;
    pthis.plugin_close = NULL;
    pthis.plugin_next = NULL;
    pthis.plugin_sweep = NULL;
    pthis.plugin_output = NULL;
    pthis.plugin_results = NULL;
    pthis.plugin_usage = NULL;
    pthis.plugin_status = NULL;
    pthis.plugin_last_error = NULL;
    my_errno = PLUGIN_UNLOADED;
    my_file = NULL;
    my_fp = NULL;

    return PLUGIN_UNLOADED;
}

static int
plugin_open(void *c_data, int i_data)
{
    char *file_opts;            // pointer to additional opts after filename
    char *append_or_overwrite = NULL;   // indicates append or overwrite
                                        // option
    int filename_len = i_data;  // set length assuming no further options


    pthis.selected = true;      // we were selected, whether the right
                                // arguments are present or not

    if(!c_data || i_data == 0)
        return PLUGIN_NOFILE;

    file_opts = strchr(c_data, ',');

    if(file_opts != NULL)
    {
        *file_opts++ = 0;       // NULL delimit previous string, and advance
        // 
        // if there was a comma, then don't include the comma and the 
        // characters after it as part of the filename
        // 
        filename_len = strlen(c_data);

        // 
        // if an 'a' is found, then set the appropriate char option for fopen
        // other options may be filtered out with further if statements
        // 
        if(*file_opts == 'a')
            append_or_overwrite = strdup("a");
        else
            append_or_overwrite = strdup("w+");
    }
    else
        append_or_overwrite = strdup("w+");

    // check to see if my_file points to prev allocated data?
    my_file = strndup(c_data, filename_len);
#ifdef DEBUG
    printf("fopen(\"%s\", \"%s\");\n", my_file, append_or_overwrite);
#endif
    my_fp = fopen(my_file, append_or_overwrite);
    free(append_or_overwrite);
    if(my_fp == NULL)
    {
        *(pthis.err) = PLUGIN_NOFILE;
        mthis->err_ptrs->err_msg("Could not open file");
        pthis.ready = false;
        return PLUGIN_NOFILE;
    }

    pthis.ready = true;
    return PLUGIN_OK;
}

static int
plugin_close(void)
{
    fprintf(my_fp, "</sonar-scan>\n");
    fclose(my_fp);
    free(my_file);
    return PLUGIN_OK;
}

static int
plugin_output(char *msg, va_list * ap)
{
    struct tm *tm;
    char buf[60];
    time_t now;
    int tries = 0;

    /* sanity checks */
    if(!pthis.ready)
    {
        *(pthis.err) = PLUGIN_NOPEN;
        mthis->err_ptrs->err_msg("output plugin not ready");
        return PLUGIN_NOPEN;
    }
    while(!my_fp && tries < 3)
    {
        *(pthis.err) = PLUGIN_NOFILE;
        mthis->err_ptrs->err_msg("output plugin not ready, retrying");
        my_fp = fopen(my_file, "a");    /* reopen it instead of dying */
        /* return PLUGIN_NOFILE; */
        tries++;
    }
    if(!my_fp)
        return PLUGIN_NOFILE;   /* we tried */
    if(!header)
    {
        now = time(NULL);
        if(!(tm = localtime(&now)))
            mthis->err_ptrs->err_sys("could not get localtime\n");
        if(strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M %Z", tm) <= 0)
            mthis->err_ptrs->err_sys("Could not format localtime\n");
        fprintf(my_fp, "<?xml version=\"1.0\" ?>\n");
        fprintf(my_fp, "<!-- %s %s initiated at %s as %s -->\n", PACKAGE,
                VERSION, buf, mthis->opts.sCommandLine);
        fprintf(my_fp, "<sonar-scan>\n");
        header++;
    }
    fprintf(my_fp, "<!-- ");
    vfprintf(my_fp, msg, *ap);
    fprintf(my_fp, " -->\n");
    return PLUGIN_OK;
}

static int
null_result(target_t * result)
{
    /* as of right now, there is no way to report dead hosts */
    fprintf(my_fp,
            "\t<deadhost hostname=\"unreported-in-this-version\" />\n");
    return PLUGIN_OK;
}

static int
plugin_results(target_t * result, unsigned int num_hosts)
{
    struct sockaddr_in *sa = NULL;
    struct sockaddr_in6 *sa6 = NULL;
    char ip6sa_format[] = "%x%x%x%x:%x%x%x%x:%x%x%x%x:%x%x%x%x\0";
    char buf[256];
    /* struct sockaddr_un *un; */

    int i, j;

    for(j = 0; j < num_hosts; j++)
    {
        assert(result != NULL);
        if(result->addrinfo == NULL)
        {
            null_result(result);
            continue;
        }

        memset(buf, 0, 256);
        fprintf(my_fp,
                "\t<host hostname=\"%s\" rtt=\"%0.2f\""
                "ttl=\"%d\" mask=\"%d\"",
                result->addrinfo->ai_canonname,
                result->stats.rtt_avg, result->stats.ttl, result->netmask);
        switch (result->addrinfo->ai_family)
        {
        case PF_INET:
            sa = (struct sockaddr_in *) result->addrinfo->ai_addr;
            fprintf(my_fp,
                    " address=\"%s\" addr-type=\"%s\">\n",
                    inet_ntoa(sa->sin_addr), "ipv4");
            break;
        case PF_INET6:
            sa6 = (struct sockaddr_in6 *) result->addrinfo->ai_addr;
            snprintf(buf, 255, ip6sa_format,
                     sa6->sin6_addr.s6_addr[0],
                     sa6->sin6_addr.s6_addr[1],
                     sa6->sin6_addr.s6_addr[2],
                     sa6->sin6_addr.s6_addr[3],
                     sa6->sin6_addr.s6_addr[4],
                     sa6->sin6_addr.s6_addr[5],
                     sa6->sin6_addr.s6_addr[6],
                     sa6->sin6_addr.s6_addr[7],
                     sa6->sin6_addr.s6_addr[8],
                     sa6->sin6_addr.s6_addr[9],
                     sa6->sin6_addr.s6_addr[10],
                     sa6->sin6_addr.s6_addr[11],
                     sa6->sin6_addr.s6_addr[12],
                     sa6->sin6_addr.s6_addr[13],
                     sa6->sin6_addr.s6_addr[14], sa6->sin6_addr.s6_addr[15]);
            fprintf(my_fp,
                    " address=\"%s\" addr-type=\"%s\">\n", buf, "ipv6");
            break;
            /* case PF_UNIX: un = (struct sockaddr_un*)
             * result->ainfo->ai_addr; fprintf(my_fp, " address=\"%s\"
             * addr-type=\"%s\">\n", un->sun_path, "unix"); break; */
        default:
            fprintf(my_fp,
                    " address=\"%s\" addr-type=\"%s\">\n",
                    "unknown-type", "unknown");
        };

        /* print the port info */
        for(i = 0; i < 65536; i++)
        {
            switch (result->ports[i])
            {
            case PORT_OPENED:
                fprintf(my_fp,
                        "\t\t<port number=\"%d\" " "state=\"open\"/>\n", i);
                break;
            case PORT_CLOSED:
                fprintf(my_fp,
                        "\t\t<port number=\"%d\" " "state=\"closed\"/>\n", i);
                break;
            case PORT_FILTERED:
                fprintf(my_fp,
                        "\t\t<port number=\"%d\" "
                        "state=\"filtered\"/>\n", i);
                break;
            case PORT_UNFILTERED:
                fprintf(my_fp,
                        "\t\t<port number=\"%d\" "
                        "state=\"unfiltered\"/>\n", i);
                break;
            }
        }
        fprintf(my_fp,
                "\t\t<stats sent=\"%d\" recvd=\"%d\" dropped=\"%d\" ",
                result->stats.packets_sent, result->stats.packets_recvd,
                result->stats.packets_sent - result->stats.packets_recvd);
        fprintf(my_fp,
                " rtt_min=\"%.2f\" rtt_max=\"%.2f\" rtt_avg=\"%.2f\" "
                "/>\n", result->stats.rtt_min, result->stats.rtt_max,
                result->stats.rtt_avg);
        fprintf(my_fp, "\t</host>\n");
        fflush(my_fp);
        result = result->next;
    }
    return PLUGIN_OK;
}

static void
plugin_usage(void)
{
    assert(mthis != NULL);
    assert(mthis->output != NULL);
    mthis->output("\n XML output plugin\n");
    mthis->output("  Usage: -oX,<filename>,[o/a]\n");
    mthis->output("   [o/a]: either 'o' or 'a', for overwrite, or append\n");
    mthis->output("          overwrite is default\n");
    /* exit(-1); */
}

static void
plugin_status(void)
{
    assert(mthis != NULL);
    assert(mthis->output != NULL);
    if(pthis.ready && pthis.selected)
        mthis->output("standard XML output loaded::(%s)\n", my_file);
}

static int
plugin_last_error(void)
{
    return my_errno;
}
