/***************************************************************************
   network_icmp.c  -  ICMP network plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: network_icmp.c,v 1.47 2004/05/06 20:11:09 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file network_icmp.c
  * File contains ICMP plugin functions
  * This file is a good example of the plugin API.
  * Use it as an example for your own plugin development.
  */

#include <assert.h>
#include <pthread.h>
#include <errno.h>
#include "network_icmp.h"
#include "mem.h"
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

static const char rcsid[] =
    "\100$ autosec/sonar: $Id: network_icmp.c,v 1.47 2004/05/06 20:11:09 red0x Exp $";

static unsigned char done = 0;

/// my last error
static uint16_t my_errno;
/// my socket
static int my_socket,
/// my sequence number
  seq_no, seq_no2,
/// lowest sequence number
  low_seq,
/// my ICMP type
  my_type,
/// my ICMP code
  my_code;

/// ready for use
static bool ready;
/// payload file name
static char *payload_file;
/// Timestamp Time zone
static struct timezone tz;
//static unsigned int num_hosts;

/// Plugin's plugin_t structure, built by plugin_init
static plugin_t pthis;
/// Sonar's sonar_t structure, built by sonar, passed to this plugin
static sonar_t *mthis;
static icmp_options_t my_options;


static int icmp_end(void);
static int icmp_leadin(void *c_data, int i_data);
static int icmp_close(void);
static int icmp_sweep(target_t * targets);
static void icmp_usage(void);
static void icmp_status(void);
static int icmp_ping(target_t * target);
static void *recv_thread(void *arg);
static unsigned short checksum(unsigned short *buffer, int size);
static int icmp_last_error(void);
static void inline error(char *x, int y);
static int resptype(int type);
static const char *respstr(int type);
static const char *unreach_codes(int code);
static RETSIGTYPE sig_handler(int i);
static void tvsub(struct timeval *out, struct timeval *in);

/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */
void *
plugin_init(void *in_data)
{
    assert(in_data != NULL);
    // useless reference, to avoid a warning
    if(rcsid) ;

    // this is assumed
    mthis = (sonar_t *) in_data;
    if(mthis->api < 6)
        mthis->err_ptrs->err_msg("Plugin API must be 6 or higher, update sonar");

    srand(getpid() ^ (uid_t) time(NULL));
    my_errno = PLUGIN_LOADED;
    my_socket = -1;

    seq_no = (int) (256.0 * rand() / (RAND_MAX + 0.0));
    payload_file = NULL;
    pthis.version = 6;          /* set API version 5 */

    // setup my plugin stuff
    pthis.type = NETWORK_PLUGIN;
    pthis.err = (int16_t *) & my_errno;
    pthis.ready = false;

    pthis.plugin_char = 'I';
    pthis.shortname = "icmp scan";
    pthis.longname = "ICMP Scanning Plugin";

    pthis.plugin_init = &plugin_init;
    pthis.plugin_destroy = &icmp_end;
    pthis.plugin_open = &icmp_leadin;
    pthis.plugin_close = &icmp_close;
    pthis.plugin_next = NULL;
    pthis.plugin_sweep = &icmp_sweep;
    pthis.plugin_sweep = &icmp_sweep;
    pthis.plugin_output = NULL;
    pthis.plugin_results = NULL;
    pthis.plugin_usage = &icmp_usage;
    pthis.plugin_status = &icmp_status;
    pthis.plugin_last_error = &icmp_last_error;
    ready = false;

    // num_hosts = 0;
    my_type = ICMP_ECHO;
    my_code = 0;

    return &pthis;
}

/** @fn icmp_end(void)
  * Plugin Unload function
  * @return PLUGIN_UNLOADED
  */
static int
icmp_end(void)
{
    my_errno = PLUGIN_UNLOADED;
    if(my_socket != -1 && my_socket != 0)
    {
        close(my_socket);
        my_socket = -1;
    }

    seq_no = 0;

    if(payload_file != NULL)
    {
        free(payload_file);
        payload_file = NULL;
    }

    pthis.type = NETWORK_PLUGIN;
    pthis.err = (int16_t *) & my_errno;
    pthis.ready = false;
    pthis.plugin_init = NULL;
    pthis.plugin_destroy = NULL;
    pthis.plugin_open = NULL;
    pthis.plugin_close = NULL;
    pthis.plugin_next = NULL;
    pthis.plugin_sweep = NULL;
    pthis.plugin_output = NULL;
    pthis.plugin_results = NULL;
    pthis.plugin_usage = NULL;
    pthis.plugin_status = NULL;
    pthis.plugin_last_error = NULL;
    ready = false;

    mthis = NULL;
    return PLUGIN_UNLOADED;
}

static void
icmp_arg_parse(char *sArgs, int arglen)
{
    char *string = sArgs;
    int i;

    while(sArgs)
    {
#ifdef DEBUG
        printf("while(\'%s\')\n", sArgs);
#endif
        i = strlen(sArgs);
        if(strncmp(sArgs, "payload=", (i < 8) ? i : 8) == 0)
        {
            /* we have an option, so set it */
            string = strchr(sArgs, ',');
            sArgs = strchr(sArgs, '=');
            if(!sArgs)
                continue;
            sArgs++;
            if(string)
                *string = 0;    /* make the comma a NULL delimiter */
#ifdef DEBUG
            printf("imcp_arg_parse: payload = \'%s\':%d\n", sArgs,
                   strlen(sArgs));
#endif
            mthis->memput(&payload_file, sArgs, strlen(sArgs));
            if(string)
                *string = ',';
        }
        else if(strncmp(sArgs, "type=", (i < 5) ? i : 5) == 0)
        {
            int mlen;
            string = strchr(sArgs, ',');
            sArgs = strchr(sArgs, '=');
            if(!sArgs)
                continue;
            sArgs++;
            if(string)
                *string = 0;
#ifdef DEBUG
            printf("icmp_arg_parse: type = \'%s\'\n", sArgs);
#endif
            mlen = strlen(sArgs);
            my_type = -1;
            if(strncmp(sArgs, "echo", (mlen < 4) ? mlen : 4) == 0)

                my_type = ICMP_ECHO;
            else if(strncmp(sArgs, "du", (mlen < 2) ? mlen : 2) == 0)
                my_type = ICMP_UNREACH;
            else if(strncmp(sArgs, "mask", (mlen < 4) ? mlen : 4) == 0)
                my_type = ICMP_MASKREQ;
            else if(strncmp(sArgs, "red", (mlen < 3) ? mlen : 3) == 0)
                my_type = ICMP_REDIRECT;
            else if(strncmp(sArgs, "ts", (mlen < 2) ? mlen : 2) == 0)
                my_type = ICMP_TSTAMP;
            else if(strncmp(sArgs, "info", (mlen < 4) ? mlen : 4) == 0)
                my_type = ICMP_IREQ;
            else if(strncmp(sArgs, "param", (mlen < 5) ? mlen : 5) == 0)
                my_type = ICMP_PARAMPROB;
            else if(strncmp(sArgs, "help", (mlen < 4) ? mlen : 4) == 0)
            {
                mthis->err_ptrs->err_quit("Currently supported types are:\n"
                         "\t echo - echo request\n"
                         "\t du - dest unreachable\n"
                         "\t mask - mask request\n"
                         "\t red - redirect\n"
                         "\t ts - timestamp request\n"
                         "\t info - info request\n"
                         "\t param - param problem\n");
            }
            else
            {
                mthis->err_ptrs->err_quit("type not supported\n"
                         "Currently supported types are:\n"
                         "\t echo - echo request\n"
                         "\t du - dest unreachable\n"
                         "\t mask - mask request\n"
                         "\t red - redirect\n"
                         "\t ts - timestamp request\n"
                         "\t info - info request\n"
                         "\t param - param problem\n");
            }
            if(!ICMP_INFOTYPE(my_type)
               && my_type != ICMP_PARAMPROB
               && my_type != ICMP_UNREACH && my_type != ICMP_REDIRECT)
            {
                mthis->err_ptrs->err_quit("type not supported");
            }
            if(string)
                *string = ',';
        }
        else if(strncmp(sArgs, "code=", (i < 5) ? i : 5) == 0)
        {
            string = strchr(sArgs, ',');
            sArgs = strchr(sArgs, '=');
            if(!sArgs)
                continue;
            sArgs++;
            if(string)
                *string = 0;
            my_code = atoi(sArgs);
#ifdef DEBUG
            printf("icmp_arg_parse: code = \'%d\'\n", my_code);
#endif
            if(string)
                *string = ',';
        }
        else if(strncmp(sArgs, "ttl=", (i < 4) ? i : 4) == 0)
        {
            string = strchr(sArgs, ',');
            sArgs = strchr(sArgs, '=');
            if(!sArgs)
                continue;
            sArgs++;
            if(string)
                *string = 0;
            my_options.ttl = atoi(sArgs);
#ifdef DEBUG
            printf("icmp_arg_parse: ttl = \'%d\'\n", my_options.ttl);
#endif
            if(string)
                *string = ',';
        }
        else if(strncmp(sArgs, "size=", (i < 5) ? i : 5) == 0)
        {
            string = strchr(sArgs, ',');
            sArgs = strchr(sArgs, '=');
            if(!sArgs)
                continue;
            sArgs++;
            if(string)
                *string = 0;
            my_options.size = atoi(sArgs);
#ifdef DEBUG
            printf("icmp_arg_parse: size = \'%d\'\n", my_options.size);
#endif
            if(string)
                *string = ',';
        }
        else if(strncmp(sArgs, "debug", (i < 5) ? i : 5) == 0)
        {
            my_options.debug = 1;
#ifdef DEBUG
            printf("icmp_arg_parse: debug mode\n");
#endif
        }
        sArgs = strchr(sArgs, ',');
        if(sArgs)
            sArgs++;
    }
}

/** @fn icmp_leadin(void *c_data, int i_data)
  * Plugin Start function
  * @param c_data Not used yet
  * @param i_data Not used yet
  * @return PLUGIN_OK
  */
static int
icmp_leadin(void *c_data, int i_data)
{
    // load socket stuff here
    uid_t uid;
    struct linger l;

    assert(mthis != NULL);
    assert(mthis->usage != NULL);
    assert(mthis->getprivs != NULL);
    assert(mthis->dropprivs != NULL);

    if(c_data && i_data > 0)
        icmp_arg_parse(c_data, i_data);

    if(mthis->opts.iPostDelay == -1)
        mthis->opts.iPostDelay = 0;     // default Post Delay of (NONE)
    // setuid root section of code
    {
        uid = mthis->getprivs();
        // root section of code
        // get me a RAW ICMP socket
        my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
        if(my_socket == -1)
        {
            mthis->err_ptrs->err_ret("could not get raw socket\n");
            exit(-1);
        }

        if(my_options.ttl > 0)
            if(setsockopt
               (my_socket, SOL_IP, IP_TTL, &(my_options.ttl),
                sizeof(my_options.ttl)) < 0)
            {
                mthis->err_ptrs->err_ret("could not set TTL");
            }

        l.l_onoff = 1;
        l.l_linger = mthis->opts.iTimeout;
        if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER,
                      (struct linger *) &l, sizeof(struct linger)) < 0)
        {
            mthis->err_ptrs->err_ret("could not set timeout");
            // perror("setsockopt");
        }

        if(my_options.debug)
        {
            l.l_onoff = 1;
            if(setsockopt(my_socket, SOL_SOCKET, SO_DEBUG,
                          (int *) &(l.l_onoff), sizeof(int)) < 0)
            {
                mthis->err_ptrs->err_ret("could not set debug mode");
            }
        }
        if(mthis->opts.bBroadcast)
        {
            l.l_onoff = 1;
            if(setsockopt(my_socket, SOL_SOCKET, SO_BROADCAST,
                          (int *) &l.l_onoff, sizeof(int)) < 0)
            {
                mthis->err_ptrs->err_ret("could not set broadcast mode");
            }
        }
        mthis->dropprivs(uid);
    }
    // back to normalcy
    ready = true;
    pthis.ready = true;
    seq_no = 1 + (int) (70.0 * rand() / (RAND_MAX + 1.0));
    return PLUGIN_OK;
}

/** @fn icmp_close(void)
  * Plugin destroy function
  * Call this before icmp_end
  * @return PLUGIN_OK
  */
static int
icmp_close(void)
{
    close(my_socket);
    my_socket = -1;
    ready = false;
    payload_file = NULL;
    pthis.ready = false;
    return PLUGIN_OK;
}

/** @fn icmp_sweep(target_t *targets)
  * Plugin Sweep Function
  * Runs a network sweep (this is for network plugins only).
  * Recompiled to make one pass only!!!
  * @param targets Pointer to linked list of targets
  * @return PLUGIN_ERROR on error, PLUGIN_OK on ok
  */
static int
icmp_sweep(target_t * targets)
{
    unsigned int counter = 0;
    pthread_t tid;
    pthread_attr_t attr;
    struct recv_t recd;
    int timeout;

    assert(targets != NULL);

    // just make sure we, and sonar, are ready for processing
    if(!ready || !pthis.ready)
    {
        mthis->err_ptrs->err_msg("plugin not ready\n");
        return PLUGIN_ERROR;
    }

    timeout = mthis->opts.iTimeout;
    if(timeout <= 0)
        timeout = 5;            // default timeout of 3 seconds

    pthread_attr_init(&attr);
    memset(&recd, 0, sizeof(struct recv_t));

    recd.socket = my_socket;

    low_seq = recd.start = clock();

    /* maximum timeout value */
    recd.max = timeout;
    recd.hosts = mthis->num_targets;

    /* add time for a per host delay */
    if(mthis->opts.iHostDelay > 0)
        recd.max += mthis->opts.iHostDelay;
    /* recd.max now contains the maximum ammount of seconds to wait per host */

    done = 0;
    if(pthread_create(&tid, &attr, recv_thread, &recd) != 0)
        mthis->err_ptrs->err_sys("could not create thread");

    while(targets != NULL && counter < mthis->num_targets)
    {
        /* we have a broadcast target, without an ip-range */
        if(targets->broadcast == 1 && targets->extra_hosts == 0)
            /* add to the timeout value */
            recd.hosts += 255;
        /* we have an ip-range here */
        if(targets->extra_hosts > 0)
        {
            /* increas the timeout */
            recd.hosts += targets->extra_hosts;
#ifdef DEBUG
            fprintf(stderr, "%s: increasing recd.hosts by %d\n",
                    targets->entry, targets->extra_hosts);
#endif
            if(mthis->opts.iVerbose > 1)
                mthis->err_ptrs->err_msg("Increasing timeout, netmask detected\n");
        }
        if(icmp_ping(targets) == PLUGIN_ERROR)
        {
            return PLUGIN_ERROR;
        }
        /* usleep(10); *//* wait a while */

        targets = targets->next;
        counter++;
        /* sleep after this host (allows more dest unreachables to come back) 
         */
        if(counter < mthis->num_targets && mthis->opts.iHostDelay > 0)
            usleep(mthis->opts.iHostDelay * 1000);
    }
    done = 1;
    /* wait for the thread to exit */
    pthread_join(tid, NULL);

    if(mthis->opts.iPostDelay)
        usleep(mthis->opts.iPostDelay * 1000);
    return PLUGIN_OK;
}

/** @fn icmp_status(void)
  * Plugin Status hook function
  * Prints out status information right before the sweep begins
  */
static void
icmp_status(void)
{
    if(pthis.ready)
    {
        mthis->output("ICMP sweep: %d host(s)::\n", mthis->num_targets);
        mthis->output("\tType: %s Code: %d (non-verified)\n",
                      respstr(my_type), my_code);
        if(mthis->opts.iPostDelay == 0)
            mthis->output("\tno post-delay::\n");
        else
            mthis->output("\tpost-delay %ds::\n", mthis->opts.iPostDelay);
        if(payload_file != NULL)
            mthis->output("\tpayload: %s::\n", payload_file);
        else
            mthis->output("\tno payload::\n");

        if(my_options.ttl != 0)
            mthis->output("\tTTL %d::\n", my_options.ttl);
        else
            mthis->output("\tdefault TTL::\n");
        if(my_options.size != 0)
            mthis->output("\tSize %d::\n", my_options.size);
        else
            mthis->output("\tRandom Size::\n");
    }
}

/** @fn icmp_usage(void)
  * Plugin Usage hook function
  * Prints out plugin usage information. All options preceded by -p are passed
  * off to plugins, so make your usage reflect that.  However, when your
  	arguments
  * get to your plugin's args function, the p will be stripped.
  */
static void
icmp_usage(void)
{
    assert(mthis->output != NULL);

    mthis->output("\n ICMP Scan plugin: requires root priveleges\n");
    mthis->
        output
        ("  Usage: -sI,[payload=<filename>,type=[echo],code=[int],ttl=[int],size=[int]]\n");
    mthis->
        output
        ("\t  payload: append contents of file to payload of ICMP packet\n");
    mthis->output("\t  type: set ICMP type (type=help for a list)\n");
    mthis->output("\t  code: set ICMP code\n");
    mthis->output("\t  ttl:  set IP TTL\n");
    mthis->output("\t  size: set packet size\n");
    mthis->
        output
        ("  Example: sonar -sI,payload=/etc/passwd,type=echo,code=0,ttl=255,size=64\n");
    mthis->output("                target.example.com\n");
    /* Do not exit, so other plugins can ouput their usage too! exit(-1); */
}

/** @fn icmp_ping(target_t *target)
  * Plugin Ping function
  * Sends out an ICMP ECHO_REQUEST packet
  * @param target pointer to a host_list_t structure containing current host to
  	ping
  * @return PLUGIN_OK on ok, PLUGIN_ERROR otherwise
  */
static int
icmp_ping(target_t * target)
{
    struct icmp *icmp_hdr;
    int total_size = sizeof(struct icmp), dlen = 0;
    int bwrote;
    uint32_t addr, one = 1;
    unsigned int j;
    // struct timezone tz;
    // register struct timeval *tp;
    char *buf, *datapart;
    struct sockaddr_in *sa;
    struct in_addr ia;
    struct timeval *tv;

    assert(target != NULL);
    assert(target->entry != NULL);
    assert(target->addrinfo != NULL);
    assert(target->addrinfo->ai_addr != NULL);
    assert(my_socket != 0);
    assert(my_socket != -1);

    srand(getpid() ^ (uid_t) time(NULL));

    /* set the packet size to specified */
    if(my_options.size != 0)
        total_size += my_options.size;
    // or pick a random size 64 to 290
    else
        total_size += 64 + (int) (290.0 * rand() / (RAND_MAX + 64.0));


    /* adding timeval structure to end of packet */
    total_size += sizeof(struct timeval);

    for(j = 0; j <= target->extra_hosts; j++)
    {
        /* get the right address */
        switch (target->addrinfo->ai_family)
        {
        case PF_INET:
            sa = (struct sockaddr_in *) target->addrinfo->ai_addr;
            ia = sa->sin_addr;
            addr = ntohl(ia.s_addr);
            if(j > 0)
            {
                addr++;
                ia.s_addr = htonl(addr);
            }
            sa->sin_addr = ia;
            /* broadcast address, set bcast mode */
            if(((addr & 0x000000ff) == 0) || ((addr & 0x000000ff) == 255))
            {
                if(one)
                {
                    if(setsockopt(my_socket, SOL_SOCKET, SO_BROADCAST,
                                  (int *) &one, sizeof(int)) < 0)
                    {
                        mthis->err_ptrs->err_ret("could not set broadcast mode");
                    }
                    one = 0;
                }
            }
            break;
        default:
            mthis->err_ptrs->err_quit("address family not supported yet\n");
        }
        /* how much of the packet is data? */
        dlen = total_size - sizeof(struct icmp) - sizeof(struct timeval);
        buf = (char *) Malloc(total_size);
        memset(buf, 0, total_size);

        icmp_hdr = (struct icmp *) buf;
        icmp_hdr->icmp_type = my_type;
        icmp_hdr->icmp_code = my_code;
        icmp_hdr->icmp_id = (unsigned short) getpid();
        icmp_hdr->icmp_cksum = 0;
        /* random sequence number, to make fingerprinting this tool harder */
        low_seq = 1 + (int) (25.0 * rand() / (RAND_MAX + 1.0));
        icmp_hdr->icmp_seq = seq_no2 = low_seq + (int) (25.0 * rand()
                                                        / (RAND_MAX + 1.0));
        /* time value included in packet */
        tv = (struct timeval *) (buf + sizeof(struct icmp));

        datapart = buf + sizeof(struct icmp) + sizeof(struct timeval);
        /* no payload */
        if(!payload_file)
        {
            /* Pattern Option Removed. --rd */
            /* do we want a pattern? */
            /* saved for later, just in case if(mthis->pCon.sPattern) { int
             * i, len = strlen(mthis->pCon.sPattern); int num = dlen / len;
             * for(i = 0; i < num; i++) { strncpy(datapart,
             * mthis->pCon.sPattern, len); datapart += len; } } */
            /* or random garbage */

            /* else */
            {
                FILE *rp;
                rp = fopen("/dev/urandom", "r");
                if(!rp)
                {
#ifdef DEBUG
                    fprintf(stderr,
                            "network_icmp.c: could not "
                            "open /dev/urandom\n");
#endif
                    rp = fopen("/dev/random", "r");
                    if(!rp)
                    {
#ifdef DEBUG
                        fprintf(stderr,
                                "network_icmp.c: could not "
                                "open /dev/random\n");
#endif
                        memset(datapart,
                               (int) (256.0 * rand() /
                                      (RAND_MAX + 0.0)), dlen);
                    }
                }

                /* if we didn't set it already */

                if(*datapart == 0)
                {
                    fread(datapart, dlen, 1, rp);
                    fclose(rp);
                }
            }

        }
        /* we've got a payload */
        else
        {
            unsigned int len = 0, cur = 0;
            FILE *payload;
            payload = fopen(payload_file, "r");
            if(!payload)
            {
                *(pthis.err) = PLUGIN_NOFILE;
                mthis->err_ptrs->err_ret("could not open payload file");
                free(buf);
                return PLUGIN_ERROR;
            }
            fseek(payload, 0, SEEK_END);
            len = (unsigned int) ((ftell(payload) > dlen) ? dlen :
                                  ftell(payload));
            fseek(payload, 0, SEEK_SET);
            while(!feof(payload) && cur < len)
            {
                int c = fgetc(payload);
                datapart[cur++] = c;
            }
        }

        /* Changed to per-packet based timestamp */
        gettimeofday(tv, &(tz));

        ((struct icmp *) buf)->icmp_cksum =
            checksum((unsigned short *) buf, total_size);
        bwrote = sendto(my_socket, buf, total_size, 0,
                        target->addrinfo->ai_addr,
                        target->addrinfo->ai_addrlen);
        /* sizeof(struct sockaddr)); */
        free(buf);
        if(bwrote == -1)
        {
            mthis->err_ptrs->err_ret("sendto failed");
            // perror("sendto");
            return PLUGIN_ERROR;
        }
        if(bwrote < (int) total_size)
        {
            mthis->err_ptrs->err_msg("wrote %d bytes\n", bwrote);
        }
        target->stats.packets_sent++;
    }
    return PLUGIN_OK;
}

/** @fn recv_thread(void *arg)
  * Plugin Receive thread
  * Looks for ICMP ECHO_REPLY packets
  * @param arg Pointer to recv_t structure
  * @return This thread returns NULL when its done.
  */
static void *
recv_thread(void *arg)
{
    fd_set rfds;
    struct ip *iphdr;
    struct icmp *icmphdr;
    unsigned short iphdrlen;
    double triptime;
    unsigned long num_res = 0;
    struct timeval *tv, tr;
    int bread, ret, len;
    char buf[MAX_PACKET], *adr = NULL;
    struct recv_t *rec = (struct recv_t *) arg;
    int fromlen, my_socket = rec->socket, result;
    struct sockaddr_in from;
    static struct addrinfo *ai = NULL;
    int psuedo = false;
    target_t *my_host;

    /* signals, baby */
    signal(SIGTERM, sig_handler);
    signal(SIGINT, sig_handler);
    signal(SIGHUP, sig_handler);
    signal(SIGSEGV, sig_handler);

    rec->responses = 0;

    while(done == 0
          ||
          ((clock() <= (rec->max * CLOCKS_PER_SEC * rec->hosts) + rec->start)
           && rec->responses < rec->hosts))
    {
        tr.tv_sec = 0;
        tr.tv_usec = 0;
        my_host = NULL;

        FD_ZERO(&rfds);
        FD_SET(my_socket, &rfds);

        ret = select(my_socket + 1, &rfds, NULL, NULL, &tr);
        if(ret == -1)
        {
            *(pthis.err) = errno;
            mthis->err_ptrs->err_ret("could not select socket\n");
            return NULL;        // PLUGIN_ERROR;
        }
        if(ret == 0)
        {
            continue;
        }
        if(psuedo == true)
        {
            assert(ai != NULL);
            psuedo = false;
            free(ai);
        }
        memset(&from, 0, sizeof(struct sockaddr_in));
        memset(buf, 0, MAX_PACKET);

        bread = recvfrom(my_socket, buf, MAX_PACKET, 0,
                         (struct sockaddr *) &(from),
                         (socklen_t *) & (fromlen));
        if(bread == -1)
        {
            mthis->err_ptrs->err_ret("recvfrom");
            // perror("recvfrom");
            continue;
        }

        iphdr = (struct ip *) buf;
        iphdrlen = iphdr->ip_hl * 4;    /* number of 32-bit words *4 = bytes */

        /* for some reason, this gets crapped on (prolly has to do with ipv6
         * wrapping */
        if(from.sin_addr.s_addr == 0)
            from.sin_addr.s_addr = iphdr->ip_src.s_addr;

        if(bread < iphdrlen + ICMP_MIN + sizeof(struct timeval))
            mthis->message("Too few bytes from %s\n",
                           inet_ntoa(from.sin_addr));

        icmphdr = (struct icmp *) (buf + iphdrlen);
        /* timestamp of time this packet was sent */
        tv = (struct timeval *) (buf + iphdrlen + sizeof(struct icmp));

        adr = Malloc(25);
        memset(adr, 0, 25);
        strncpy(adr, inet_ntoa(iphdr->ip_src), 24);
        if((ai = mthis->resolve(adr, NULL, 0, 0)) == NULL)
        {
            len = strlen(adr);
            psuedo = true;
            ai = (struct addrinfo *) Malloc(sizeof(struct addrinfo));
            memset(ai, 0, sizeof(struct addrinfo));
            ai->ai_family = PF_INET;
            ai->ai_protocol = IPPROTO_TCP;
            ai->ai_addrlen = sizeof(struct sockaddr_in);
            ai->ai_addr =
                (struct sockaddr *) Malloc(sizeof(struct sockaddr_in));
            memset(ai->ai_addr, 0, sizeof(struct sockaddr_in));
            memcpy(ai->ai_addr, &(from.sin_addr), sizeof(struct sockaddr_in));
            ai->ai_canonname = (char *) Malloc(len + 1);
            memset(ai->ai_canonname, 0, len + 1);
            strncpy(ai->ai_canonname, adr, len);
        }
        free(adr);

        if((my_host = mthis->find_target(mthis, ai)) == NULL)
        {
            continue;
        }

        gettimeofday(&tr, &tz);

        /* how long since this packet has been gone */
        tvsub(&tr, tv);
        triptime = tr.tv_sec * 1000.0 + tr.tv_usec / 1000.0;

        if(icmphdr->icmp_type != resptype(my_type))
        {
            if(icmphdr->icmp_type == ICMP_DEST_UNREACH)
            {
                mthis->message("%s: %s(%d) \n",
                               inet_ntoa(icmphdr->icmp_ip.ip_dst),
                               unreach_codes(icmphdr->icmp_code),
                               icmphdr->icmp_code);
                rec->responses++;

                break;
            }
            else if(mthis->opts.iVerbose > 0)
            {
                mthis->message("%s[%s]: non-%s recv'd - type: %s "
                               "code: ", ai->ai_canonname,
                               inet_ntoa(from.sin_addr),
                               respstr(resptype(my_type)),
                               respstr(icmphdr->icmp_type));

                mthis->output("id: %d -> %s\n",
                              icmphdr->icmp_id, ai->ai_canonname);
            }
            continue;
        }

        if(triptime < my_host->stats.rtt_min || my_host->stats.rtt_min == 0)
            my_host->stats.rtt_min = triptime;
        if(triptime > my_host->stats.rtt_max || my_host->stats.rtt_max == 0)
            my_host->stats.rtt_max = triptime;
        if(my_host->stats.rtt_avg == 0)
            my_host->stats.rtt_avg = triptime;
        else
            my_host->stats.rtt_avg =
                ((((double) my_host->stats.packets_recvd - 1)
                  * my_host->stats.rtt_avg) + triptime)
                / (double) my_host->stats.packets_recvd;
        if(my_host->stats.rtt_avg < 0)
            my_host->stats.rtt_avg *= -1;

        my_host->stats.ttl = iphdr->ip_ttl;
        my_host->stats.packets_recvd++;

        num_res++;

        mthis->message
            ("%s[%s]\tonline: icmp_seq = %d. range: %.2f ms ttl: "
             " %d size: %d\n",
             ai->ai_canonname, inet_ntoa(from.sin_addr),
             icmphdr->icmp_seq, triptime, iphdr->ip_ttl,
             ntohs(iphdr->ip_len));

        /* 
         * this'll be useful later - it's going to be a plugin option instead
         * of a global option
         if(mthis->pCon.prog != NULL)
         {
         int retv;
         retv = system(mthis->pCon.prog);
         if(WIFSIGNALED(retv) &&
         (WTERMSIG(retv) == SIGINT
         || WTERMSIG(retv) == SIGQUIT))
         {
         done = 1;
         pthread_exit(NULL);
         }
         }
         */
        rec->responses++;
        continue;
    }

    result = rec->hosts - rec->responses;
    if(result > 0)
        mthis->output("no response(s) from %d hosts\n", result);
    if(result < 0)
        mthis->output("subnet broadcast detected: extra response(s) "
                      "from %d hosts\n", ~result + 1);
    return NULL;
}

/**
  * @fn checksum(unsigned short *buffer, int size)
  * Compute an ICMP checksum
  * @param buffer Pointer to data
  * @param size Size of data
  * @return Checksum of data
  */
static unsigned short
checksum(unsigned short *buffer, int size)
{
    unsigned long cksum = 0;

    while(size > 1)
    {
        cksum += *buffer++;
        size -= sizeof(unsigned short);
    }

    if(size)
    {
        cksum += *(unsigned char *) buffer;
    }

    cksum = (cksum >> 16) + (cksum & 0xffff);
    cksum += (cksum >> 16);
    return (unsigned short) (~cksum);
}

/**
  * @fn icmp_last_error()
  * Return last error code
  * @return Last error code
*/
static int
icmp_last_error(void)
{
    return my_errno;
}

/**
  * @fn error(char *x, int y)
  * Display and error and set the error code
  * @param x Message to display
  * @param y Error code to set
*/
static void inline
error(char *x, int y)
{
    mthis->message("%s: ERROR %s\n", __FILE__, x);
    my_errno = y;
}

/**
  * @fn resptype(int type)
  * Return the code for the response type 
  * @param type ICMP type sent
  * @return Expected response type
*/
static int
resptype(int type)
{
    switch (type)
    {
    case ICMP_ROUTERADVERT:
        return ICMP_ROUTERSOLICIT;
    case ICMP_ROUTERSOLICIT:
        return ICMP_ROUTERADVERT;
    case ICMP_TSTAMP:
        return ICMP_TSTAMPREPLY;
    case ICMP_IREQ:
        return ICMP_IREQREPLY;
    case ICMP_MASKREQ:
        return ICMP_MASKREPLY;
    case ICMP_ECHO:
        return ICMP_ECHOREPLY;
    }
    return -1;
}

/**
  * @fn respstr(int type)
  * Return the name for the response type 
  * @param type ICMP type sent
  * @return Expected response type in a string
*/
static const char *
respstr(int type)
{
    switch (type)
    {
    case ICMP_UNREACH:
        return "dest unreachable";
    case ICMP_SOURCEQUENCH:
        return "source quench";
    case ICMP_ROUTERADVERT:
        return "router advert";
    case ICMP_ROUTERSOLICIT:
        return "router solicit";
    case ICMP_TIMXCEED:
        return "time exceed";
    case ICMP_PARAMPROB:
        return "param problem";
    case ICMP_TSTAMP:
        return "timestamp request";
    case ICMP_TSTAMPREPLY:
        return "timestamp reply";
    case ICMP_IREQ:
        return "info request";
    case ICMP_IREQREPLY:
        return "info reply";
    case ICMP_MASKREQ:
        return "mask request";
    case ICMP_MASKREPLY:
        return "mask reply";
    case ICMP_ECHO:
        return "echo request";
    case ICMP_ECHOREPLY:
        return "echo reply";
    }
    return "unknown";
}

/**
  * @fn unreach_codes(int code)
  * Return the name of the dest_unreach code 
  * @param code ICMP code number
  * @return ICMP code in a string form
*/
static const char *
unreach_codes(int code)
{
    switch (code)
    {
        // unreach codes:
    case ICMP_UNREACH_NET:
        return "net unreachable";
    case ICMP_UNREACH_HOST:
        return "host unreachable";
    case ICMP_UNREACH_PROTOCOL:
        return "protocol not supported";
    case ICMP_UNREACH_PORT:
        return "port unreachable";
    case ICMP_UNREACH_NEEDFRAG:
        return "need fragmentation";
    case ICMP_UNREACH_SRCFAIL:
        return "source-route failed";
    case ICMP_UNREACH_NET_UNKNOWN:
        return "network unknown";
    case ICMP_UNREACH_HOST_UNKNOWN:
        return "host unknown";
    case ICMP_UNREACH_ISOLATED:
        return "host isolated";
    case ICMP_UNREACH_NET_PROHIB:
        return "net denied";
    case ICMP_UNREACH_HOST_PROHIB:
        return "host denied";
    case ICMP_UNREACH_TOSNET:
        return "bad TOS for net";
    case ICMP_UNREACH_TOSHOST:
        return "bad TOS for host";
    case ICMP_UNREACH_FILTER_PROHIB:
        return "filter prohibited";
    case ICMP_UNREACH_HOST_PRECEDENCE:
        return "host precedence violation";
    case ICMP_UNREACH_PRECEDENCE_CUTOFF:
        return "precedence cutoff";
    }
    return "unknown";
}

static RETSIGTYPE
sig_handler(int i)
{
    switch (i)
    {
    case SIGSEGV:
        mthis->err_ptrs->err_msg("oops! ACK scan seg faulted.\n");
        abort();
        break;
    }
    /* just close the thread and be done with it */
    pthread_exit(NULL);
}

/* out = out - in */
static void
tvsub(struct timeval *out, struct timeval *in)
{
    if((out->tv_usec -= in->tv_usec) < 0)
    {
        out->tv_sec--;
        out->tv_usec += 1000000;
    }
    out->tv_sec -= in->tv_sec;
}
