/***************************************************************************
    rfc793.c  -  TCP ACK scan plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: rfc793.c,v 1.53 2004/05/06 20:11:09 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file rfc793.c
  * Contains a rfc 793 behavior check (TCP ACK->RST ping)
  */

#include <assert.h>
#include <pthread.h>
#include "rfc793.h"
#include "mem.h"
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

#include <errno.h>

static const char rcsid[] =
    "\100$ autosec/sonar: $Id: rfc793.c,v 1.53 2004/05/06 20:11:09 red0x Exp $";

static int my_socket, err_socket;
static unsigned char done = 0;
static plugin_t pthis;
static sonar_t *mthis;
static int ready;

static uint16_t my_errno;
static struct timezone tz;
static struct timeval tp;
static struct rfc_options op;

static int plugin_destroy(void);
static void rfc793_options_parse(void *, int);
static int plugin_open(void *c_data, int i_data);
static int plugin_close(void);
static void plugin_usage(void);
static void plugin_status(void);
static int plugin_last_error(void);
static int rfc793_sweep(target_t * targets);
static void tvsub(struct timeval *out, struct timeval *in);
static int rfc793_send(target_t * target);
static int recv_thread(target_t *target, int my_socket, int err_socket, unsigned int timeout);
static void print_packet(struct tcphdr *tcp);


static int in_cksum(unsigned short *addr, int len);
static void add_tcphdr(unsigned char *pkt, struct net_tuple *nt,
                       uint8_t flags);
static int send_ipv4(int sox, u_char * pkt, size_t len, struct net_tuple *nt);
static const char *unreach_codes(int code);

/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */
void *
plugin_init(void *in_data)
{
    assert(in_data != NULL);

    /* useless reference, to avoid a warning */
    if(rcsid) ;

    /* this is assumed */
    mthis = (sonar_t *) in_data;
    if(mthis->api < 6)
        mthis->err_ptrs->err_quit("Plugin API must be 6 or higher, update sonar");

    pthis.type = NETWORK_PLUGIN;
    pthis.err = &my_errno;
    pthis.ready = true;
    pthis.version = 6;          /* set API version 5 */
    ready = false;

    pthis.plugin_char = 'A';
    pthis.shortname = "ack scan";
    pthis.longname = "ACK scanning plugin";

    pthis.plugin_init = &plugin_init;
    pthis.plugin_destroy = &plugin_destroy;
    pthis.plugin_open = &plugin_open;
    pthis.plugin_close = &plugin_close;
    pthis.plugin_next = NULL;
    pthis.plugin_sweep = &rfc793_sweep;
    pthis.plugin_output = NULL;
    pthis.plugin_results = NULL;
    pthis.plugin_usage = &plugin_usage;
    pthis.plugin_status = &plugin_status;
    pthis.plugin_last_error = &plugin_last_error;

    my_errno = PLUGIN_LOADED;
    my_socket = -1;

    return &pthis;
}


static int
plugin_destroy(void)
{
    mthis = NULL;
    if(my_socket > 0)
    {
        close(my_socket);
        my_socket = -1;
    }

    if(op.prog != NULL)
        free(op.prog);
    pthis.type = 0;
    pthis.err = NULL;
    pthis.ready = false;
    ready = false;
    pthis.plugin_init = NULL;
    pthis.plugin_destroy = NULL;
    pthis.plugin_open = NULL;
    pthis.plugin_close = NULL;
    pthis.plugin_next = NULL;
    pthis.plugin_sweep = NULL;
    pthis.plugin_output = NULL;
    pthis.plugin_results = NULL;
    pthis.plugin_usage = NULL;
    pthis.plugin_status = NULL;
    pthis.plugin_last_error = NULL;
    my_errno = PLUGIN_UNLOADED;
    my_socket = -1;

    return PLUGIN_UNLOADED;
}

static void
rfc793_options_parse(void *arguments, int len)
{
    char *endval, *arg, *tmp;
    int i, vallen;
    arg = (char *) arguments;
    op.srcPrt = -1;

    while(arg)
    {
        i = strlen(arg);

        if((i >= 3 && strncasecmp(arg, "ttl", (i < 3) ? i : 3)) == 0)
        {
            arg += 3;
            while(*arg == '=' || *arg == ' ')
                arg++;
            op.ttl = strtol(arg, NULL, 0xA);    // :)
        }

        if((i >= 5 && strncasecmp(arg, "debug", (i < 5) ? i : 5)) == 0)
        {
            op.debug = 1;
        }

        if((i >= 6 && strncasecmp(arg, "srcprt", (i < 6) ? i : 6)) == 0)
        {
            arg += 6;
            while(*arg == '=' || *arg == ' ')
                arg++;
            op.srcPrt = strtol(arg, NULL, 0xA);
            if(errno == EINVAL)
                op.srcPrt = -1;
        }

        if((i >= 4 && strncasecmp(arg, "prog", (i < 4) ? i : 4)) == 0)
        {
            arg += 4;
            while(*arg == '=' || *arg == ' ')
                arg++;

            // find out how long the command is
            endval = strchr(arg, ',');
            if(endval == NULL)
            {
                vallen = strlen(arg);
            }
            else
            {
                for(vallen = 0, tmp = arg; tmp != endval && *tmp != '\0';
                    tmp++, vallen++) ;
            }
            op.prog = (char *) malloc(vallen * sizeof(char));
            strncpy(op.prog, arg, vallen);
            *(op.prog + vallen) = '\0';
        }

        arg = strchr(arg, ',');
        if(arg)
            arg++;
    }
}




static int
plugin_open(void *c_data, int i_data)
{
    // load socket stuff here
    uid_t uid;
    struct linger l;

    assert(mthis != NULL);
    assert(mthis->usage != NULL);
    assert(mthis->getprivs != NULL);
    assert(mthis->dropprivs != NULL);

    rfc793_options_parse(c_data, i_data);

    // setuid root section of code
    {
        uid = mthis->getprivs();
        // root section of code
        // get me RAW TCP and ICMP sockets
        my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_TCP);
        if(my_socket == -1)
            mthis->err_ptrs->err_sys("could not create raw TCP socket");
        err_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
        if(err_socket == -1)
            mthis->err_ptrs->err_sys("could not create raw ICMP socket");

        if(op.ttl != 0)
        {
            if(setsockopt(my_socket, SOL_IP, IP_TTL, &(op.ttl),
                          sizeof(op.ttl)) < 0)
            {
                *(pthis.err) = PLUGIN_ERROR;
                mthis->err_ptrs->err_ret("could not set TTL");
            }
        }

        l.l_onoff = 0;
        if(mthis->opts.iTimeout == -1)
            l.l_linger = 5;
        else
            l.l_linger = mthis->opts.iTimeout;
        if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER, (struct linger *) &l,
                      sizeof(struct linger)) < 0)
        {
            *(pthis.err) = PLUGIN_ERROR;
            mthis->err_ptrs->err_ret("could not set timeout");
            // perror("setsockopt");
            // mthis->usage("could not set linger time on TCP socket");
        }
        if(setsockopt(err_socket, SOL_SOCKET, SO_LINGER, (struct linger *) &l,
                      sizeof(struct linger)) < 0)
        {
            *(pthis.err) = PLUGIN_ERROR;
            mthis->err_ptrs->err_ret("could not set icmp timeout");
            // perror("setsockopt");
            // mthis->usage("could not set linger time on ICMP socket");
        }
        if(op.debug)
        {
            l.l_onoff = 1;
            if(setsockopt
               (my_socket, SOL_SOCKET, SO_DEBUG, (int *) &(l.l_onoff),
                sizeof(int)) < 0)
            {
                *(pthis.err) = PLUGIN_ERROR;
                mthis->err_ptrs->err_ret("could not set debug mode");
            }

        }
        if(mthis->opts.bBroadcast)
        {
            printf("setting broadcast\n");
            l.l_onoff = 1;
            if(setsockopt(my_socket, SOL_SOCKET, SO_BROADCAST,
                          (int *) &(l.l_onoff), sizeof(int)) < 0)
            {
                *(pthis.err) = PLUGIN_ERROR;
                mthis->err_ptrs->err_ret("could not set broadcast mode");
                // perror("setsockopt");
                // exit(-1);
            }
        }

        mthis->dropprivs(uid);
    }
    // back to normalcy

    ready = true;
    pthis.ready = true;
    return PLUGIN_OK;
}

static int
plugin_close(void)
{
    close(my_socket);
    pthis.ready = false;

    return PLUGIN_OK;
}

static void
plugin_usage(void)
{
    assert(mthis != NULL);
    assert(mthis->output != NULL);
    mthis->output("\n ACK Scan plugin: requires root priveleges\n");
    mthis->
        output
        ("  Usage: -sA,[ttl=<num>,srcprt=<port>,prog=<cmdline>] hostname:port/mask\n");
    mthis->output("  \tttl: set IP TTL\n");
    mthis->output("  \tsrcprt: set TCP source port\n");
    mthis->
        output
        ("  \tprog: set command line to execute if host(s) is/are online\n");
    mthis->
        output
        ("  If no port is specified, this will scan *all* ports in /etc/services\n");
    mthis->
        output
        ("    Example: sonar -sA,ttl=64,srcprt=53,prog=\"echo hi\" target.example.com:80/24\n");
}

static void
plugin_status(void)
{
    assert(mthis != NULL);
    assert(mthis->output != NULL);
    if(pthis.ready && ready)
    {
        mthis->output("ACK Scan: %d host(s)::\n", mthis->num_targets);
        if(mthis->opts.iPostDelay == 0)
            mthis->output("\tno post-delay::\n");
        else
            mthis->output("\tpost-delay %ds::\n", mthis->opts.iPostDelay);
        if(op.ttl != 0)
            mthis->output("\tTTL %d::\n", op.ttl);
        else
            mthis->output("\tdefault TTL::\n");
        if(op.srcPrt != -1)
            mthis->output("\tSource port %d::\n", op.srcPrt);
        else
            mthis->output("\tRandom source port\n");
    }
}

static int
plugin_last_error(void)
{
    return my_errno;
}

static int
rfc793_sweep(target_t * targets)
{
    unsigned int counter = 0, i;
    int timeout, n, j;
    unsigned int num_responses = 0, num_targets = 0, num_hosts = 0, num_ports = 0;
	unsigned int filt = 0, unfilt = 0, unreach = 0, k = 0;
	int result = 0;

    assert(targets != NULL);

    // just make sure we, and sonar, are ready for processing
    if(!pthis.ready || !ready)
    {
        mthis->err_ptrs->err_msg("plugin not ready\n");
        return PLUGIN_ERROR;
    }

    timeout = mthis->opts.iTimeout;
    if(timeout <= 0)
        timeout = 5;            // default timeout of 5 seconds

    while(targets != NULL && counter < mthis->num_targets)
    {
		num_hosts++;
		j = 1;
        for(i = 0; i < 65536; i++)
        {
            /* stupid hax0r tr1ckz */
            if(targets->ports[i] == PORT_SCAN)
            {
				num_targets++;
				num_ports++;
                targets->ports[i] = PORT_SCANNING;
                /* scan this one port */
                if(rfc793_send(targets) == PLUGIN_ERROR)
                {
                    /* XXX: report an error */
                    return PLUGIN_ERROR;
                }
				k = 0;
				while(k++ <= targets->extra_hosts)
				{
					/* try to receive, but if we get interrupted, its a timeout */
					n = recv_thread(targets, my_socket, err_socket, timeout);
					switch(n)
					{
						case RFC_NORESP:
							break;
						case RFC_FILTERED:
							filt++;num_responses++;
							break;
						case RFC_UNFILT:
							unfilt++;num_responses++;
							break;
						case RFC_UNREACH:
							if(j == 1)
							{
								unreach++;
								j = 0;
							}
							num_responses++;
							break;
						case PLUGIN_ERROR:
							if(mthis->opts.iVerbose > 1)
								mthis->err_ptrs->err_msg("Error in recv_thread\n");
							break;
					}
					
					if(errno == EINTR)
					{
						/* we timed out */
						if(mthis->opts.iVerbose > 1)
							mthis->err_ptrs->err_msg("ACK Scan: Network Timeout");
					}
				}
            }
        }
        /* sleep after this host (allows more dest unreachables to
        * come back) */
        if(counter < mthis->num_targets && mthis->opts.iHostDelay > 0)
            usleep(mthis->opts.iHostDelay * 1000);
        counter++;
		
        targets = targets->next;
    }

    if(mthis->opts.iPostDelay)
    {
        /* sleep this many milliseconds after this scan */
        usleep(mthis->opts.iPostDelay * 1000);
    }
	
	if(mthis->opts.iVerbose > 2)
    {
        result = num_targets - num_responses;
        if(result > 0)
            mthis->output("no response(s) from %d hosts\n", result);
       if(result < 0 && num_ports == 1)
            mthis->output("subnet broadcast detected: extra response(s) from "
                          "%d hosts\n", ~result + 1);
        mthis->message("ACK Scan complete:\n");
        mthis->message("  Scanned %d hosts/ %d ports, %d hosts online, %d "
                       "filtered, %d unfiltered, %d hosts unreachable\n",
                       num_hosts,
                       num_ports,
                       num_hosts - unreach, filt, unfilt, unreach);
    }
    if(mthis->opts.iVerbose <= 1 && unfilt && mthis->opts.iVerbose > 0)
        mthis->
            output
            ("\tThe remaining %d ports that were in state \'unfiltered\' were\n\tnot printed, if you want to see them, use '-v2'\n",
             unfilt);
    return PLUGIN_OK;
}

/* @fn rfc793_send(target_t *target)
 * Send a specially crafted TCP ACK packet to first entry in *target
 * @param target Filled target_t pointer
 * @return PLUGIN_OK if all went well
 * @todo Timestamp each packet individually
 */
static int
rfc793_send(target_t * target)
{
    unsigned char *packet = NULL;
    struct addrinfo *ai = NULL;
    struct net_tuple nt;
    struct sockaddr_in *sa = NULL;
    struct in_addr ia;
    char name[256];
    int bwrote, i;
    unsigned int j;
    int total_size = sizeof(struct tcphdr);
    // XXX:
    // was my_ports meant for something? if not it's not doing anything
    // useful
    int my_ports[65547];
    uint32_t addr, one = 1;

    assert(target != NULL);
    assert(target->entry != NULL);
    assert(target->addrinfo != NULL);
    assert(target->addrinfo->ai_addr != NULL);
    assert(my_socket != 0);
    assert(my_socket != -1);
    assert(mthis->output != NULL);

    srand(getpid() ^ (uid_t) time(NULL));
    gethostname(name, 255);
    ai = mthis->resolve(name, NULL, PF_INET, 0);
    memcpy(my_ports, target->ports, 65537 * sizeof(int));
    for(j = 0; j <= target->extra_hosts; j++)
    {
        packet = (unsigned char *) Malloc(total_size);
        memset(packet, 0, total_size);
        memset(name, 0, 256);

        if(!ai)
            return PLUGIN_RESOLVE;
        while(ai->ai_family != PF_INET)
        {
            fprintf(stderr, "non ipv4 address:\n");
            ai = ai->ai_next;
            if(!ai)
            {
                fprintf(stderr, "no more addresses");
                return PLUGIN_RESOLVE;
            }
        }
        sa = (struct sockaddr_in *) ai->ai_addr;
        memcpy(&nt.src, &(sa->sin_addr), sizeof(struct in_addr));

        if(op.srcPrt != -1)
            nt.sport = htons(op.srcPrt);
        else
            nt.sport = htons(1 + (int) (65535.0 * rand() / (RAND_MAX + 1.0)));

        switch (target->addrinfo->ai_family)
        {
        case PF_INET:
            sa = (struct sockaddr_in *) target->addrinfo->ai_addr;
            ia = sa->sin_addr;
            addr = ntohl(ia.s_addr);
            if(j > 0)
            {
                addr++;
                ia.s_addr = htonl(addr);
            }
            sa->sin_addr = ia;
            /* broadcast address, set bcast mode */
            if(((addr & 0x000000ff) == 0) || ((addr & 0x000000ff) == 255))
            {
                if(one)
                {
                    if(setsockopt(my_socket, SOL_SOCKET, SO_BROADCAST,
                                  (int *) &one, sizeof(int)) < 0)
                    {
                        mthis->err_ptrs->err_ret("could not set broadcast mode");
                    }
                    one = 0;
                }
            }
            nt.dst = ia.s_addr;
            break;
        default:
            mthis->err_ptrs->err_quit("address family not supported yet\n");
        }

        //nt.dport = -1;
        for(i = 0; i < 65536; i++)
        {
            /* lame hax0r tr1ck (1 = not scanned, 2 = scanning, 3 = scanned) */

            if(target->ports[i] == PORT_SCANNING)
            {
                target->ports[i] = PORT_SCANNED;
                nt.dport = htons(i);
            }
        }

        add_tcphdr(packet, &nt, TH_ACK);
        // packet is ready to send 

        // time stamp our packet, for round trip time
        gettimeofday(&(tp), &(tz));

        bwrote = send_ipv4(my_socket, packet, total_size, &nt);
        if(bwrote == -1)
        {
            mthis->err_ptrs->err_ret("send_ipv4 failed");
            // perror("send_ipv4");
            return PLUGIN_ERROR;
        }
        if(bwrote < (int) total_size)
        {
            mthis->err_ptrs->err_msg("wrote %d bytes\n", bwrote);
        }
        target->stats.packets_sent++;
        free(packet);
		memcpy(target->ports, my_ports, 65537 * sizeof(int));
    }

    return PLUGIN_OK;
}

static int
recv_thread(target_t *target, int my_socket, int err_socket, unsigned int timeout)
{
    fd_set rfds;
    struct ip *iphdr;
    struct icmp *icmphdr;
    struct tcphdr *tcp = NULL;
    unsigned short iphdrlen;
    struct timeval tv;
    int bread, ret;
    double triptime;
    char buf[MAX_PACKET], *adr = NULL;
    int fromlen;
    struct sockaddr_in from;
    static struct addrinfo *ai = NULL;
    int psuedo = false;
    target_t *my_host;

    assert(my_socket != -1 && my_socket != 0);
    assert(err_socket != -1 && err_socket != 0);
    assert(mthis->resolve != NULL);
    
	
    while(1)
    {
        /* dont do anything up here, or we'll eat CPU */
        my_host = NULL;
		tv.tv_usec = 0;
		tv.tv_sec = timeout;
		
        FD_ZERO(&rfds);
        FD_SET(my_socket, &rfds);
        FD_SET(err_socket, &rfds);
        ret = select(((my_socket > err_socket) ? my_socket + 1 :
                      err_socket + 1), &rfds, NULL, NULL, &tv);
        if(ret == -1)
        {
            *(pthis.err) = errno;

            mthis->err_ptrs->err_ret("could not select socket\n");
            return PLUGIN_ERROR;
        }
        if(ret == 0)    /* timeout */
            return RFC_NORESP;
        if(psuedo == true)
        {
            assert(ai != NULL);
            psuedo = false;
            free(ai);
        }

        memset(&from, 0, sizeof(struct sockaddr_in));
        memset(buf, 0, MAX_PACKET);
        if(FD_ISSET(my_socket, &rfds))
        {
            /* read our packet */
            bread = recvfrom(my_socket, buf, MAX_PACKET, 0,
                             (struct sockaddr *) &(from),
                             (socklen_t *) & (fromlen));
            if(bread == -1)
            {
                *(pthis.err) = errno;
                mthis->err_ptrs->err_ret("recvfrom error");
                continue;       // return PLUGIN_ERROR;
            }
            /* XXX: timestamping should be on a per-packet basis */
            gettimeofday(&tv, &(tz));
            tvsub(&tv, &tp);
            iphdr = (struct ip *) buf;
            iphdrlen = iphdr->ip_hl * 4;        // number of 32-bit words *4
                                                // = bytes

            if(bread < iphdrlen + sizeof(struct tcphdr))
                mthis->message("Too few bytes from %s\n",
                               inet_ntoa(from.sin_addr));

            tcp = (struct tcphdr *) (buf + iphdrlen);

            if(from.sin_addr.s_addr == 0)
                from.sin_addr.s_addr = iphdr->ip_src.s_addr;

            adr = Malloc(25);
            memset(adr, 0, 25);
            strncpy(adr, inet_ntoa(iphdr->ip_src), 24);
            if((ai = mthis->resolve(adr, NULL, 0, 0)) == NULL)
            {
                int len = strlen(adr);
                psuedo = true;
                ai = (struct addrinfo *) Malloc(sizeof(struct addrinfo));
                memset(ai, 0, sizeof(struct addrinfo));
                ai->ai_family = PF_INET;
                ai->ai_protocol = IPPROTO_TCP;
                ai->ai_addrlen = sizeof(struct sockaddr_in);
                ai->ai_addr =
                    (struct sockaddr *) Malloc(sizeof(struct sockaddr_in));
                memset(ai->ai_addr, 0, sizeof(struct sockaddr_in));
                memcpy(ai->ai_addr, &(from.sin_addr),
                       sizeof(struct sockaddr_in));
                ai->ai_canonname = (char *) Malloc(len + 1);
                memset(ai->ai_canonname, 0, len + 1);
                strncpy(ai->ai_canonname, adr, len);
            }
            free(adr);

            if((my_host = mthis->find_target(mthis, ai)) == NULL)
            {
				/* not from one of our targets, so ignore it */
                continue;
            }

            // we get something other than a RST, it should be filtered
            triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);
            if((tcp->th_flags & TH_RST))
            {
                my_host->ports[ntohs(tcp->th_sport)] = PORT_UNFILTERED;
                if(triptime < my_host->stats.rtt_min
                   || my_host->stats.rtt_min == 0)
                    my_host->stats.rtt_min = triptime;
                if(triptime > my_host->stats.rtt_max
                   || my_host->stats.rtt_max == 0)
                    my_host->stats.rtt_max = triptime;
                if(my_host->stats.rtt_avg == 0)
                    my_host->stats.rtt_avg = triptime;
                else
                    my_host->stats.rtt_avg =
                        ((((double) my_host->stats.packets_recvd - 1)
                          * my_host->stats.rtt_avg) + triptime) /
                        (double) my_host->stats.packets_recvd;
                if(my_host->stats.rtt_avg < 0)
                    my_host->stats.rtt_avg *= -1;
                my_host->stats.ttl = iphdr->ip_ttl;
                my_host->stats.packets_recvd++;

		/*
		 * PHARKAS
		 * Call your function like this
		 */
		//mthis->foobar(char*string_to_expand, tcp->th_sport, my_host);

                if(mthis->opts.iVerbose > 1)
                    mthis->
                        message
                        ("%s[%s]:%d online & UNfiltered: range: %.2f ms "
                         "ttl: %d size: %d\n",
                         my_host->addrinfo->ai_canonname,
                         inet_ntoa(from.sin_addr), ntohs(tcp->th_sport),
                         triptime, iphdr->ip_ttl, ntohs(iphdr->ip_len));

                /* print the packet if they want it */
                if(mthis->opts.iVerbose > 3)
                    print_packet(tcp);

                if(op.prog != NULL)
                {
                    ret = system(op.prog);
                    if(WIFSIGNALED(ret) &&
                       (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT))
                    {
                        done = 1;
                        pthread_exit(NULL);
                    }
                }

                return RFC_UNFILT;
            }
			/* non-rst packet, keep looking */
            continue;
        }
        else if(FD_ISSET(err_socket, &rfds))
        {
            int len;
            bread = recvfrom(err_socket, buf, MAX_PACKET, 0,
                             (struct sockaddr *) &(from),
                             (socklen_t *) & (fromlen));
            if(bread == -1)
            {
                // perror("recvfrom");
                mthis->err_ptrs->err_ret("recvfrom error");
                return PLUGIN_ERROR;
            }

            gettimeofday(&tv, &(tz));

            tvsub(&tv, &tp);
            triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);

            iphdr = (struct ip *) buf;
            iphdrlen = iphdr->ip_hl * 4;        // number of 32-bit words *4
                                                // = bytes

            if(bread < iphdrlen + sizeof(struct icmp))
                mthis->message("Too few bytes from %s\n",
                               inet_ntoa(from.sin_addr));

            icmphdr = (struct icmp *) (buf + iphdrlen);

            tcp = (struct tcphdr *) (icmphdr + sizeof(struct icmp));

            if(icmphdr->icmp_type != ICMP_DEST_UNREACH)
            {
                continue;       // return PLUGIN_OK;
                // Parse non-unreachable ICMP returns
            }
            /* change the address information to reflect where this packet
             * was headed */
            from.sin_addr = icmphdr->icmp_ip.ip_dst;

            adr = Malloc(25);
            memset(adr, 0, 25);
            strncpy(adr, inet_ntoa(from.sin_addr), 24);
            if((ai = mthis->resolve(adr, NULL, 0, 0)) == NULL)
            {
                len = strlen(adr);
                psuedo = true;
                ai = (struct addrinfo *) Malloc(sizeof(struct addrinfo));
                memset(ai, 0, sizeof(struct addrinfo));
                ai->ai_family = PF_INET;
                ai->ai_protocol = IPPROTO_TCP;
                ai->ai_addrlen = sizeof(struct sockaddr_in);
                ai->ai_addr = (struct sockaddr *)
                    Malloc(sizeof(struct sockaddr_in));
                memcpy(ai->ai_addr, &(from.sin_addr),
                       sizeof(struct sockaddr_in));
                ai->ai_canonname = (char *) Malloc(len + 1);
                memset(ai->ai_canonname, 0, len + 1);
                strncpy(ai->ai_canonname, adr, len);
            }
            free(adr);

            if((my_host = mthis->find_target(mthis, ai)) == NULL)
            {
				/* not one of our targets, keep looking */
                continue;
            }

            switch (icmphdr->icmp_code)
            {
            case ICMP_UNREACH_PORT:
                assert(ai != NULL);
                my_host->ports[ntohs(tcp->th_dport)] = PORT_FILTERED;
                if(triptime < my_host->stats.rtt_min
                   || my_host->stats.rtt_min == 0)
                    my_host->stats.rtt_min = triptime;
                if(triptime > my_host->stats.rtt_max
                   || my_host->stats.rtt_max == 0)
                    my_host->stats.rtt_max = triptime;
                if(my_host->stats.rtt_avg == 0)
                    my_host->stats.rtt_avg = triptime;
                else
                    my_host->stats.rtt_avg =
                        ((((double) my_host->stats.packets_recvd - 1) *
                          my_host->stats.rtt_avg) + triptime) /
                        (double) my_host->stats.packets_recvd;
                if(my_host->stats.rtt_avg < 0)
                    my_host->stats.rtt_avg *= -1;

                my_host->stats.ttl = iphdr->ip_ttl;
                my_host->stats.packets_recvd++;

                if(mthis->opts.iVerbose > 0)
                    mthis->message("%s[%s]:%d\tonline & filtered: range: "
                                   "%.2f ms ttl: %d size: %d\n",
                                   my_host->addrinfo->ai_canonname,
                                   inet_ntoa(from.sin_addr),
                                   ntohs(tcp->th_dport),
                                   triptime, iphdr->ip_ttl,
                                   ntohs(iphdr->ip_len));
                return RFC_FILTERED;
            case ICMP_UNREACH_HOST:
                assert(ai != NULL);
                if(mthis->opts.iVerbose > 0)
                    mthis->message("%s[%s] host unreachable\n",
                                   ai->ai_canonname,
                                   inet_ntoa(from.sin_addr),
                                   triptime,
                                   iphdr->ip_ttl, ntohs(iphdr->ip_len));
                return RFC_UNREACH;
            default:
                assert(ai != NULL);
                if(mthis->opts.iVerbose > 0)
                    mthis->message("%s[%s]: destination unreachable: %s\n",
                                   ai->ai_canonname,
                                   inet_ntoa(from.sin_addr),
                                   unreach_codes(icmphdr->icmp_code));
                return RFC_UNREACH;
            }
            return RFC_FILTERED;
        }
    }
    
    return RFC_NORESP;
}

/**
  * @fn tvsub(register struct timeval *out, register struct timeval *in)
  * Subtract one struct timeval from another
  * @param out Pointer to destination (will be filled with out - in)
  * @param in Pointer to src
*/
static void
tvsub(struct timeval *out, struct timeval *in)
{
    if((out->tv_usec -= in->tv_usec) < 0)
    {
        out->tv_sec--;
        out->tv_usec += 1000000;
    }
    out->tv_sec -= in->tv_sec;
}

static void
print_packet(struct tcphdr *tcp)
{
    mthis->output("  sport: %d dport: %d\n", ntohs(tcp->th_sport),
                  ntohs(tcp->th_dport));
    mthis->output("  flags: (");
    if(tcp->th_flags & TH_FIN)
        mthis->output("F");
    else
        mthis->output("-");
    if(tcp->th_flags & TH_SYN)
        mthis->output("S");
    else
        mthis->output("-");
    if(tcp->th_flags & TH_RST)
        mthis->output("R");
    else
        mthis->output("-");
    if(tcp->th_flags & TH_PUSH)
        mthis->output("P");
    else
        mthis->output("-");
    if(tcp->th_flags & TH_ACK)
        mthis->output("A");
    else
        mthis->output("-");
    if(tcp->th_flags & TH_URG)
        mthis->output("U");
    else
        mthis->output("-");
    mthis->
        output(") offset: 0x%04x window: %04x seq: 0x%04x ack: 0x%04x cksum:"
               "0x%04x\n", tcp->th_off, tcp->th_win, tcp->th_seq, tcp->th_ack,
               tcp->th_sum);
}

int
in_cksum(unsigned short *addr, int len)
{
    int nleft = len;
    int sum = 0;
    u_short *w = addr;
    u_short answer = 0;


    while(nleft > 1)
    {
        sum += *w++;
        nleft -= 2;
    }


    if(nleft == 1)              /* padding */
    {
        *(u_char *) (&answer) = *(u_char *) w;
        sum += answer;
    }

    return (sum);
}


void
add_tcphdr(unsigned char *pkt, struct net_tuple *nt, uint8_t flags)
{
    struct tcphdr tcp;
    struct _fakehead fakehead;
    int sum;
    struct timeval tv;


    memset(&tcp, 0, sizeof(tcp));
    memset(&fakehead, 0, sizeof(fakehead));
    gettimeofday(&tv, &tz);

    tcp.th_dport = nt->dport;
    tcp.th_sport = nt->sport;
    fakehead.saddr = nt->src;
    fakehead.daddr = nt->dst;
    fakehead.zero = 0, fakehead.protocol = 6;
    fakehead.tot_len = htons(TCP_SIZE);
    sum = in_cksum((u_short *) & fakehead, sizeof(fakehead));
    tcp.th_off = TCP_SIZE >> 2;
    tcp.th_seq = ntohl(1 + (unsigned long) (4294967296.0 * rand() /
                                            (RAND_MAX + 1.0)));
    tcp.th_ack = ntohl(1 + (unsigned long) (4294967296.0 * rand() /
                                            (RAND_MAX + 1.0)));
    tcp.th_flags |= flags;      /* ADD the flags */
    tcp.th_win = htons((getpid() * 1024 + rand()) % 1024);
    sum += in_cksum((u_short *) & tcp, sizeof(tcp));
    tcp.th_sum = CKSUM_CARRY(sum);
    memcpy(pkt, &tcp, sizeof(tcp));
}


/* 
 * send out ipv4-packet 
 * with data 'pkt' of length 'len' 
 * returns the number of characters sent, or -1 if an error occured 
 */
int
send_ipv4(int sox, u_char * pkt, size_t len, struct net_tuple *nt)
{
    struct sockaddr_in to;

    to.sin_family = AF_INET;
    memcpy(&to.sin_addr.s_addr, &nt->dst, sizeof(u_long));


    return (sendto(sox, pkt, len, 0, (struct sockaddr *) &to, sizeof(to)));
}

/**
  * @fn unreach_codes(int code)
  * Return the name of the dest_unreach code 
  * @param code ICMP code number
  * @return ICMP code in a string form
*/
static const char *
unreach_codes(int code)
{
    switch (code)
    {
        // unreach codes:
    case ICMP_UNREACH_NET:
        return "net unreachable";
    case ICMP_UNREACH_HOST:
        return "host unreachable";
    case ICMP_UNREACH_PROTOCOL:
        return "protocol not supported";
    case ICMP_UNREACH_PORT:
        return "port unreachable";
    case ICMP_UNREACH_NEEDFRAG:
        return "need fragmentation";
    case ICMP_UNREACH_SRCFAIL:
        return "source-route failed";
    case ICMP_UNREACH_NET_UNKNOWN:
        return "network unknown";
    case ICMP_UNREACH_HOST_UNKNOWN:
        return "host unknown";
    case ICMP_UNREACH_ISOLATED:
        return "host isolated";
    case ICMP_UNREACH_NET_PROHIB:
        return "net denied";
    case ICMP_UNREACH_HOST_PROHIB:
        return "host denied";
    case ICMP_UNREACH_TOSNET:
        return "bad TOS for net";
    case ICMP_UNREACH_TOSHOST:
        return "bad TOS for host";
    case ICMP_UNREACH_FILTER_PROHIB:
        return "filter prohibited";
    case ICMP_UNREACH_HOST_PRECEDENCE:
        return "host precedence violation";
    case ICMP_UNREACH_PRECEDENCE_CUTOFF:
        return "precedence cutoff";
    }
    return "unknown";
}
