/***************************************************************************
   plugin.h  -  plugin API
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** \file plugin.h
  * File contains plugin API \n
  * Plugin API 1.6 \n
  * To make a plugin, refer to this file's plugin_t structure
  * and to the example plugins
  */
#ifndef _PLUGIN_H_
#define _PLUGIN_H_

#include "types.h"
#include <string.h>
#include <inttypes.h>
#define DEFAULT_COUNT	1
#define DEFAULT_TIMEOUT 5000
#define API_VER		6
#define API_SUPPORT 6
#define PLUGIN_HEADER	"libsonar"

/** \name Error and Success Values
  * return these from your plugin functions
 */
//@{
/// error with last function
#define PLUGIN_ERROR	-1
/// plugin unloaded
#define PLUGIN_UNLOADED -2
/// no response (for network plugins)
#define PLUGIN_NORESP	-3
/// file not found
#define PLUGIN_NOFILE	-4
/// end of file
#define PLUGIN_EOF	-5
/// no permissions
#define PLUGIN_PERM	-6
/// not open
#define PLUGIN_NOPEN	-7
/// invalid option
#define PLUGIN_INVALIDO	-8
/// invalid argument
#define PLUGIN_INVALIDA	-9
/// unable to resolve name
#define PLUGIN_RESOLVE  -10


/// last function returned ok
#define PLUGIN_OK	1       // must correspond with SONAR_OK
/// plugin loaded
#define PLUGIN_LOADED	2
/// response (for network plugins)
#define PLUGIN_RESP	3

//@}


/** \name Plugin Types
  * Just '|' these values together for your plugin type.
  * For example for an input/output plugin, do this
  * plug->type = INPUT_PLUGIN | OUTPUT_PLUGIN;
 */
//@{
/// Input plugin
#define INPUT_PLUGIN 	1
/// Output Plugin
#define OUTPUT_PLUGIN	2
/// Network Plugin
#define NETWORK_PLUGIN	4
//@}

/** \name Max Values
  * Maximum values for certain things in the plugin API
 */
//@{

/// Max input plugins
#define MAX_INP		20
/// Max output plugins
#define MAX_OUT		20
/// Max network plugins
#define MAX_NET		20
//@}

#define MAX_PORTS   65537


#define	PORT_DONTSCAN	0
#define PORT_SCAN	1
#define PORT_SCANNING	2
#define PORT_SCANNED	3
#define PORT_OPENED	4
#define PORT_CLOSED	5
#define PORT_FILTERED	6
#define PORT_UNFILTERED	7

typedef struct stats_s stats_t;

/*! \struct stats_s
	\brief Statistics structure
	
	This structure holds the statistics per target
*/
struct stats_s
{
	//! Number of packets sent
    unsigned int packets_sent;
	//! Number of packets received
    unsigned int packets_recvd;
	//! last time to live received
    unsigned short ttl;
	//! least round trip time
    double rtt_min, 
	//! max round trip time
		rtt_max, 
	//! average round trip time
		rtt_avg;
};

/*! \struct error_ptrs
	\brief Error Function Pointers
	
	This structure holds pointers to various error functions
*/
struct error_ptrs
{
	//! Print a nonfatal error related to a system call
	void (*err_ret) (const char *fmt, ...);
	//! Print a fatal error related to a system call and exit
	void (*err_sys) (const char *fmt, ...);
	//! Print a fatal error related to a system call, dump core, and exit
	void (*err_dump) (const char *fmt, ...);
	//! Print a nonfatal error unrelated to a system call.
	void (*err_msg) (const char *fmt, ...);
	//! Print a fatal error unrelated to a system call and exit
	void (*err_quit) (const char *fmt, ...);
};

typedef struct target_s target_t;

/*! \struct target_s
	\brief Target specification
	
	This structure is filled in with various information about the targets
	This structure will be used in a linked list. Targets are filled in
	by calling sonar_add_target with the user's target specification
*/
struct target_s
{
    //! Link to the next target in list
    target_t *next;
    //! Contains only the hostname (no ports or netmasks)
    char *entry;
    //! List of ports to scan (see PORT_SCAN et al. in plugin.h)
    int ports[MAX_PORTS];
    //! Address info structue, including family, protocol, and socket type (see getaddrinfo(3))
    struct addrinfo *addrinfo;
    //! Netmask (0-32)
    unsigned int netmask;
    //! Number of extra hosts involved in this target (ie, w/netmask)
    unsigned int extra_hosts;
    //! Set if this target is a broadcast address
    unsigned short broadcast;
    //! Number of ports in this host
    uint32_t num_ports;
    //! Statistics structure
    stats_t stats;
};

/*! \struct sonar_options_s
	\brief Contains useful plugin data
	
	This structure contains all the various global options for plugin(s)
  */
typedef struct sonar_options_s
{
    //! Delay after running this type of scan (advanced option)
    int iPostDelay;
    //! Delay after each host
    int iHostDelay;
    //! Timeout per host
    int iTimeout;
    //! do we have broadcast addresses in the targets?
    int bBroadcast;

    //! Verbosity level (0 to 5)
    int iVerbose;
    //! Fork into daemon mode? (0/1)
    int bFork;
    //! Command line
    char *sCommandLine;
    //! Number of scanning cycles to run
    unsigned int iCount;
    //! Random order of hosts?
    int bRandom;
} sonar_options_t;

/*! \struct plugin_s
	\brief Definition for the plugin interface
	
	Your plugin(s) should return a filled in structure from the plugin_init function.  
	Sonar will expect all the applicable function pointers to be valid.
  */
typedef struct plugin_s
{
    //! unique ID assigned at run time by sonar
    uint32_t pluginID;
    //! plugin api interface version requested, lower versions _may_ be supported
    uint8_t version;
    //! plugin type (INPUT_PLUGIN | OUTPUT_PLUGIN for an I/O plugin)
    uint8_t type;

    //! last error from this plugin
    int16_t *err;
    //! ready to use flag, not modified by sonar
    uint8_t ready:4,
    //! The plugin was selected
      selected:4;

    //! Plugin character (ie, -sI, -sA, -sL, -oL)
    char plugin_char;
    //! Short name (ie, "icmp scan", "log file output")
    char *shortname;
    //! Long name (ie, "ICMP Scanning Plugin: requires root...")
    char *longname;

    /*! \fn plugin_init (void *in_data)
		\brief Initialize this plugin
		
		Return a pointer to a filled in plugin_s structure
	*/
    void *(*plugin_init) (void *in_data);
    //! Destroy this plugin
    int (*plugin_destroy) (void);

    /*! \fn plugin_open (void *c_data, int i_data)
		\brief Open the input/output/network source
		
		For input plugins, please return the size of the buffer you use.
		Everything else should return PLUGIN_OK on success and PLUGIN_ERROR
		on error.
	*/
    int (*plugin_open) (void *c_data, int i_data);
    //! Close the input/output/network source
    int (*plugin_close) (void);
    //! Get next host, returned in hostname, only for input plugins
    int (*plugin_next) (char *hostname);

    //! Do the sweep (network plugins)
    int (*plugin_sweep) (target_t * targets);

    //! Output plugin's general message output function
    int (*plugin_output) (char *msg, va_list * ap);
    //! output plugins result output function
    int (*plugin_results) (target_t * result, unsigned int num);

    //! prints, via sonar_t->output(), line(s) to add to sonar_usage()
    void (*plugin_usage) (void);
    //! prints, via sonar_t->output(), line(s) to add to the status report
    void (*plugin_status) (void);

    //! Return the last error
    int (*plugin_last_error) (void);

} plugin_t;

/*! \struct sonar_s
	\brief Definition for the plugin interface

	This is passed to the plugin, via plugin_init, from sonar with useful
  	function pointers and information, such as what version
  	of sonar is loading the plugins, and pointers to sonar_msg(char *,...), etc.
 */
typedef struct sonar_s
{
    //! Major version number
    uint8_t ver_major,
    //! Minor version number
      ver_minor,
    //! Revision number
      revision,
	//! Plugin API version 1.X (so for 1.4, set this to 4)
      api;

    //! Pointer to a memory copier from sonar (XXX: remove)
    void (*memput) (char **dest, const char *src, const int len);

    //! Pointer to the Messsage output (ie, with a timestamp) function
    void (*message) (char *msg, ...);

    //! Resolve a hostname, usually called with the last 3 arguments NULL, and 0
    struct addrinfo *(*resolve) (const char *host, const char *serv,
                                 int family, int socktype);
    //! Pointer to sonar's output to console function (no timestamp)
    int (*output) (char *msg, ...);

    //! Pointer to usage function, for error reports
    void (*usage) (const char *);

    //! Pointer to a function to get root priveleges
    int (*getprivs) (void);
    //! Pointer to a function to drop root priveleges
    int (*dropprivs) (uid_t);
    //! find a target, for updating
    target_t *(*find_target) (struct sonar_s * this, struct addrinfo * ai);

    //! Name of file we are (for permissions checking)
    char *my_name;

    //! struct for error function pointers
    struct error_ptrs *err_ptrs;

    //! list of attack vectors
    target_t *vectors;
    //! number of targets
    unsigned int num_targets;
    //! number of ports
    unsigned int num_ports;

    //! Useful global options
    sonar_options_t opts;
} sonar_t;

/*! \struct plugin_list_s
	\brief Wrapper for plugin_t to be used as an array
*/
typedef struct plugin_list_s
{
    //! Pointer to a plugin_t structure
    plugin_t *plug;
    //! Handle to library opened with \c lt_dlopen(3)
    void *handle;
} plugin_list_t;

int plugin_start(sonar_t * mthis);
int plugin_end(void);
int plugin_loadall(sonar_t * mthis);
int plugin_killall(void);

void plugin_status_hook(void);
int plugin_command_handler(char opt, char *optarg);
void plugin_message_hook(char *msg, va_list * ap);
void clear_opts(sonar_options_t *);


typedef plugin_t *(*PINIT) (void *);
typedef int (*PDESTROY) (void);
typedef int (*POPEN) (void *, int);
typedef int (*PCLOSE) (void);
typedef int (*PNEXT) (char *);
typedef void (*PHOOK) (void);
typedef int (*PERROR) (void);
typedef int (*PSWEEP) (void *);
typedef int (*POGO) (void);

/*! \page pluginapi Plugin API Manual
    \section v6 Version 6
  This page will explain the use and programming of
  sonar plugins.  Sonar has been trimmed down, diverting most
  of the processing to plugins.  These plugins use an API, 
  or Application Programming Interface, which provides 
  for input, output, and network plugins.  That means you can
  create your own plugins to drive different kinds of scans using
  different types of plugins. Version 6 of the API is used in sonar
  1.2.2 and above. \n\n
  
  \subsection contents Contents 
   1. \ref papi \n
     &nbsp;&nbsp; a. \ref sonari \n 
     &nbsp;&nbsp;&nbsp;&nbsp; 1. \ref load \n
     &nbsp;&nbsp;&nbsp;&nbsp; 2. \ref args \n
     &nbsp;&nbsp; b. \ref pluginint \n
     &nbsp;&nbsp;&nbsp;&nbsp; 1. \ref pload \n
     &nbsp;&nbsp;&nbsp;&nbsp; 2. \ref pother \n
   2. \ref notes \n
   3. \ref argsnotes \n
   4. \ref apichanges \n
   
  \section papi The Plugin API 
  \n
  \subsection sonari Sonar Internals
  \subsubsection load Loading plugins
    Sonar will search through \c /usr/lib/ for all \c libsonar*.so
  files (see \ref plugin_loadall). Sonar calls this function when it is
  ready to load plugins.  It passes \ref plugin_loadall a pointer to a 
  filled in \ref sonar_s structure.  \ref plugin_loadall searches through
  \c /usr/lib/ for all \c libsonar*.so files.  Any file it finds, it 
  opens with dlopen(3).  If it can't open a file, it reports which file it
  couldn't open and continues looking for more \c libsonar*.so files.  \b NOTE: \b sonar
  \b will \b not \b look \b in \b the \b PATH \b or \b the \b working \b directory
  \b for \b plugins. \b It \b will \b only \b look \b in \c /usr/lib/.  \n\n
    Once it opens your plugin, it will call dlsym(3) to load your plugin's 
  \ref plugin_init function.  It is imperative that your plugin has a function called
  \ref plugin_init (See \link inp_rand.h inp_rand.h \endlink).  Sonar then calls this function,
  passing it a pointer to a complete \ref sonar_s structure (note: sonar_s and 
  sonar_t are interchangeable).  Sonar expects that your plugin_init function return
  a pointer to a completely filled out plugin_s structure. After this, sonar adds
  a handle and your plugin_t structure to an array of plugins, depending on what
  type of plugin it is.  Sonar will store a handle of your plugin, and will call 
  dlclose when exiting the program. \n\n
  
  \subsubsection args Argument Handling
    Sonar has a few main arguments, but, with all the plugins, it could have an
  infinite amount of arguments to handle.  To counter the confusion of passing every
  plugin every argument, I made all the plugin arguments follow plugin
  selection. In other words, anything after -sA, is passed to the ack scan \b
  only. This is a change from the old way of passing '-p' options.  Each option
  needs to be comma separated.  -sA,opt1,opt2,opt3 and so on.

  If your plugin has not handled that argument, sonar expects a return value of 
  \ref PLUGIN_ERROR .  If your plugin did handle the argument (or ignores it), sonar expects a return
  value of \ref PLUGIN_OK .  \n\n
  all at once.  
  With many plugins, there is an identity crisis: which plugin gets what
  arguments? I decided to go with an nmap type solution.  Each plugin
  should be designed with a character that will uniquely identify it to sonar.  For 
  example, the \link network_icmp.c network_icmp \endlink plugin's character is \c 'I'.
  So, to use the ICMP scan plugin, you would type sonar -sI. Your plugin will have its
  \ref plugin_s::selected value set to \ref true. \n\n

  \subsection pluginint Plugin Internals
  \subsubsection pload Loading

   Once sonar has dlopen(3)'ed your plugin, it calls your plugin_init function.
  Each plugin \b *must* have a function named plugin_init, and this function 
  must return a pointer to a filled int \ref plugin_s structure.  If your
  plugin is an input/output only plugin, it can leave the plugin_sweep function 
  pointer blank (See \link inp_rand.c inp_rand.c \endlink). Your function must also setup its type correctly, or the plugin
  hooks may fail.  To set your plugin's type, set your pthis.type to the correct
  bitwise ORing ('|') of the different types (See \ref INPUT_PLUGIN, \ref OUTPUT_PLUGIN,
  \ref NETWORK_PLUGIN, and \link bare.c bare.c \endlink).
  Plugins must provide all of the other functions currently in the \ref plugin_s
  structure to avoid confusion and assert errors.  

  \subsection pother Other Questions
   If you have any questions about how to do anything else with plugins, see 
   the plugin manual , \link log_file.h log_file.h \endlink , and 
   \link plugin.h plugin.h \endlink .  If you want to make a very bare plugin that
   only does a few things, just provide all the functions and have each return 
   either PLUGIN_OK or PLUGIN_ERROR (See \link log_file.c log_file.c \endlink). 
   Also, feel free to shoot me an email at
   \c red0x@users.sourceforge.net or see me on IRC at \c irc.freenode.org:#autosec.


  \section notes Notes
   As of 1.2.2, sonar has a way to separate the plugin arguments.  The old way
   of passing options, ie: the '-p' list, is no longer supported.  For example,
   to pass a TTL option to the icmp scan, you would type \c -sI,ttl=220.
   
  After all arguments are parsed, a list of hosts should remain on the 
  command line to be parsed.  If not, currently, sonar checks for input plugins ready
  to handle filling in targets. If none are found, sonar will bail.
  Once sonar reaches this section, it creates a linked list of targets
  See \ref sonar_add_target.
  \n\n
  Once all that is taken care of, sonar passes control off to the network
  plugins.  Plugins are currently designed to be run in serial, so only do
  one pass of the sweep, make your default post-delay 0, and return PLUGIN_OK once
  your network plugin is done sweeping. 
  \n\n
   Plugins handle their arguments in the plugin_open function. The previous
   unused arguments to plugin_open are now the plugin's options, and the length
   thereof.
  Sonar will call this function with the 'command line' following your plugin's
  selection character, so you can easily use a getopt
  style switch statement to parse the arguments from there.  If your
  plugin does not handle that particular argument, do not exit, but 
  return \ref PLUGIN_ERROR. This will tell sonar that you did not handle it;
  otherwise, to ignore it or if you handled it, return \ref PLUGIN_OK.  
  \n\n
  The plugin's sweep function, if its a network function, should handle
  the sweeps however it wants. I ask that you only perform one sweep at a 
  time, for parallel scanning reasons.
  \n\n
  For more information and examples, see \link network_icmp.c 
  network_icmp.c \endlink and \link network_icmp.h network_icmp.h 
  \endlink

  \subsection argsnotes Notes on Passing Arguments
   The previous method of passing arguments, '-p' is no longer used.  See the
   above \ref notes section.\n
  \n\n

  \section apichanges Changes as of 1.5
  As of API version 1.5, I have made some options global to all plugins 
  See \ref sonar_options_s. Also, \c plugin_args has
  been removed.  All options specific to your plugin will be passed to \c plugin_open.
*/

#endif // _PLUGIN_H
