/***************************************************************************
                          target.c  -  contains functions to add targets
                             -------------------
    begin                : Tues May 13 2003
    copyright            : (C) 2003 by red0x
    email                : red0x@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file target.c
  * File contains functions to add targets
  */

#include <assert.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "target.h"
#include "service.h"
#include "net.h"
#include "error.h"
#include "log.h"
#include "sonar.h"
#include "mem.h"
#include "util.h"

extern sonar_t mthis;
static int portmask[MAX_PORTS];

static const char rcsid[] =
    "\100$ autosec/sonar: $Id: target.c,v 1.18 2004/04/13 03:42:08 red0x Exp $";

static unsigned long check_netmask(sonar_t * this, target_t * target);
static int parse_target_ports(char *ports, target_t ** target);
static int load_default_ports(target_t ** target);

int
sonar_add_target(sonar_t * this, char *entry)
{
    struct addrinfo *addr;
    target_t *target, *ent;
    char *hostname;
    char *ports;
    char *mask;
    int len, i = 0;
    unsigned int num_ports = 0, num_hosts = 0;
    int imask = 32;

    assert(this != NULL);
    assert(entry != NULL);

    len = strlen(entry);

    /* copy the entry over */
    hostname = strndup(entry, len);

    ports = mask = hostname;

    // split up the hostname and ports
    while(ports[0] != ':' && i < len)
    {
        *ports++;
        i++;
    }
    i = 0;
    while(mask[0] != '/' && i < len)
    {
        *mask++;
        i++;
    }
    // null delimit the hostname;
    if(*ports == ':')
        *ports++ = 0x00;
    else
        ports = NULL;
    // otherwise, leave it alone, assume default ports
    if(*mask == '/')
        *mask++ = 0x00;
    else
        mask = NULL;
    // ports now contains ports
    // mask now contains a hostmask

    log_write(LOG_DBUG, "  hostname: %s, ports: %s mask: %s\n", hostname,
              ports, mask);

    if(mask)
    {
        imask = strtol(mask, (char **) NULL, 10);
        if(imask < 0 || imask > 32 || errno == ERANGE || errno == EINVAL)
        {
            err_msg("Bad hostmask: %d, using /32\n", imask);
            imask = 32;
        }
    }
    target = (target_t *) Malloc(sizeof(target_t));
    memset(target, 0, sizeof(target_t));

    // start filling in needed values
    target->entry = hostname;   /* strndup(entry, len); */
    target->netmask = imask;
    // parse list of ports.
    num_ports = parse_target_ports(ports, &(target));

    // call Wrapped host_serv so we dont have to check for errors.
    addr = Host_serv(hostname, NULL, 0, 0);
    assert(addr != NULL);
    target->addrinfo = addr;
    target->num_ports = num_ports;
    target->next = NULL;
    num_hosts = check_netmask(this, target);

    ent = this->vectors;
    if(!ent)
        this->vectors = target;
    else
    {
        while(ent->next != NULL)
            ent = ent->next;
        ent->next = target;
    }

    this->num_targets++;
    return SONAR_OK;
}

/* find a target in the linked list by address info */
target_t *
sonar_find_target(sonar_t * this, struct addrinfo * ai)
{
    target_t *res = NULL;
    uint32_t start, end, longtmp;
    struct sockaddr_in *sa;

    if(!ai || !this)            /* meaningless answer to a retarded question */
        return NULL;
    if(!this->vectors)
        return NULL;

    res = this->vectors;

    if(res->addrinfo)
    {
        assert(res->addrinfo->ai_addr != NULL);
        switch (res->addrinfo->ai_family)
        {
        case PF_INET:
            sa = (struct sockaddr_in *) res->addrinfo->ai_addr;
            longtmp = ntohl(sa->sin_addr.s_addr);
            start = longtmp & (unsigned long)
                (0 - (1 << (32 - res->netmask)));
            end = longtmp | (unsigned long) ((1 << (32 - res->netmask)) - 1);
            /* if its in range, we have a match */
            sa = (struct sockaddr_in *) ai->ai_addr;
            assert(sa != NULL);
            longtmp = ntohl(sa->sin_addr.s_addr);

            if(longtmp <= end && longtmp >= start)
                return res;
            break;
        default:
            log_write(LOG_DBUG, "IPv4 supported only\n");
            return NULL;
        }
        if(memcmp(res->addrinfo, ai, sizeof(struct addrinfo)) == 0)
            return res;
    }
    while(res->next)
    {
        res = res->next;
        if(res->addrinfo)
        {
            assert(res->addrinfo->ai_addr != NULL);
            switch (res->addrinfo->ai_family)
            {
            case PF_INET:
                sa = (struct sockaddr_in *) res->addrinfo->ai_addr;
                longtmp = ntohl(sa->sin_addr.s_addr);
                start = longtmp & (unsigned long)
                    (0 - (1 << (32 - res->netmask)));
                end = longtmp | (unsigned long)
                    ((1 << (32 - res->netmask)) - 1);
                /* if its in range, we have a match */
                sa = (struct sockaddr_in *) ai->ai_addr;
                assert(sa != NULL);
                longtmp = ntohl(sa->sin_addr.s_addr);

                if(longtmp <= end && longtmp >= start)
                    return res;
                break;
            default:
                log_write(LOG_DBUG, "IPv4 supported only\n");
                return NULL;
            }
            if(memcmp(res->addrinfo, ai, sizeof(struct addrinfo)) == 0)
                return res;
        }
    }
    return NULL;
}

static unsigned long
check_netmask(sonar_t * this, target_t * target)
{
    unsigned long retval = 0, longtmp;
    uint32_t start = 0, end = 0;
    struct sockaddr_in *sa;

    assert(target != NULL);
    assert(target->addrinfo != NULL);
    assert(target->addrinfo->ai_addr != NULL);
    sa = (struct sockaddr_in *) target->addrinfo->ai_addr;
    longtmp = ntohl(sa->sin_addr.s_addr);

    start = longtmp & (unsigned long) (0 - (1 << (32 - target->netmask)));
    end = longtmp | (unsigned long) ((1 << (32 - target->netmask)) - 1);
    if(start > end)
        err_msg("Invalid netmask\n");
    retval = end - start + 1;
    if(retval <= 0)
        err_quit("Invalid hostmask\n");
    log_write(LOG_DBUG, "  Netmask: %d, hosts %u\n", target->netmask, retval);
    target->extra_hosts = retval - 1;
    /* update the target to be the start of the range */
    sa->sin_addr.s_addr = htonl(start);
    if(retval == 255 || retval == 256 || (start & 0x000000ff) == 0
       || (start & 0x000000ff) == 255)
    {
        target->broadcast = 1;
        mthis.opts.bBroadcast = 1;
    }
    return retval;
}

static int
parse_target_ports(char *ports, target_t ** target)
{
    int i = 0, len;
    char *dest, *next;
    unsigned int ret = 0;

    if(!*target)
        *target = Malloc(sizeof(target_t));
    memset((*target)->ports, PORT_DONTSCAN, MAX_PORTS);
    /* get default ports from /etc/services */
    if(!ports)
        return load_default_ports(target);

    len = strlen(ports);

    // all ports
    if(strchr(ports, '*') != NULL)
    {
        log_write(LOG_DBUG, "  setting all ports to scan\n");
        for(i = 0; i < MAX_PORTS; i++)
        {
            (*target)->ports[i] = PORT_SCAN;
            ret++;
        }
        return ret;
    }

    dest = strtok(ports, ",");

    while(dest)
    {

        // check for range
        next = strchr(dest, '-');
        if(!next)               // no range
        {
            sonar_resolve_service(dest, *target);
            ret++;
        }
        else                    // has a range
        {
            // dest has beginning
            // get the last range
            *next++ = '\0';
            ret = sonar_resolve_service_r(dest, next, *target);

        }


        dest = strtok(NULL, ",");

    }
    return ret;
}

void
load_default_ports_array()
{
    struct servent *se;
    memset(portmask, PORT_DONTSCAN, MAX_PORTS);
    while((se = getservent()) != NULL)
        portmask[se->s_port] = PORT_SCAN;
}

static int
load_default_ports(target_t ** target)
{
    int i;
    assert(*target != NULL);
    (*target)->num_ports = 0;
    for(i = 0; i < MAX_PORTS; i++)
    {
        (*target)->ports[i] = portmask[i];
        if(portmask[i] == PORT_SCAN)
            (*target)->num_ports++;
    }
    endservent();
    return (*target)->num_ports;

}

#ifdef DEBUG
int
sonar_debug_targets(target_t * target)
{
    int i = 0, n = 0;
    while(target)
    {
        err_msg(" target debug[%d]:", n++);
        err_msg("  entry: %s", target->entry);
        err_msg("  hostname: %s", target->addrinfo->ai_canonname);
        err_msg("   broadcast: %d", target->broadcast);
        err_msg("   num_hosts: %d", target->extra_hosts);
        err_msg("  ports:");
        for(i = 0; i < 65536; i++)
            if(target->ports[i] == PORT_SCAN)
                printf("%d,", i);
        err_msg("");
        target = target->next;
    }

    return SONAR_OK;
}
#endif // DEBUG
